"""
Various ugly utility functions for twill.

Apart from various simple utility functions, twill's robust parsing
code is implemented in the ConfigurableParsingFactory class.
"""

import os
import re

from typing import Any, List, NamedTuple, Optional, Union, Sequence, Tuple

from requests import Response
from requests.structures import CaseInsensitiveDict
from lxml.html import (
    fromstring as html_to_tree, tostring as tree_to_html,
    CheckboxGroup, FormElement, HtmlElement, InputElement,
    MultipleSelectOptions, RadioGroup, SelectElement, TextareaElement)

try:
    import tidylib  # type: ignore
except (ImportError, OSError):
    # ImportError can be raised when PyTidyLib package is not installed
    # OSError can be raised when the HTML Tidy shared library is not installed
    tidylib = None

from . import log, twill_ext
from .errors import TwillException

__all__ = [
    'gather_filenames', 'get_equiv_refresh_interval', 'html_to_tree',
    'is_hidden_filename', 'is_twill_filename', 'print_form',
    'make_boolean', 'make_int', 'make_twill_filename',
    'run_tidy', 'tree_to_html',  'trunc', 'unique_match',
    'CheckboxGroup', 'FieldElement', 'FormElement',
    'HtmlElement', 'InputElement', 'Link', 'RadioGroup',
    'ResultWrapper', 'SelectElement', 'Singleton', 'TextareaElement',
    'UrlWithRealm', 'Response']


FieldElement = Union[
    CheckboxGroup, InputElement, RadioGroup, SelectElement, TextareaElement]


class Link(NamedTuple):
    text: str
    url: str


# Depending on the configuration, realms can be ignored
UrlWithRealm = Union[str, Tuple[str, str]]


class Singleton:
    """A mixin class to create singleton objects."""

    def __new__(cls, *args, **kwargs):
        it = cls.__dict__.get('__it__')
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        return it

    @classmethod
    def reset(cls):
        cls.__it__ = None


class ResultWrapper:
    """Deal with request results, and present them in a unified form.

    These objects are returned by browser._journey()-wrapped functions.
    """
    def __init__(self, response: Response) -> None:
        self.response = response
        self.encoding = response.encoding
        try:
            self.tree = html_to_tree(self.text)
        except ValueError:
            # may happen when there is an XML encoding declaration
            self.tree = html_to_tree(self.content)
        self.xpath = self.tree.xpath
        self._fix_forms()

    @property
    def url(self) -> str:
        """"Get the url of the result page."""
        return self.response.url

    @property
    def http_code(self) -> int:
        """Get the http status code of the result page."""
        return self.response.status_code

    @property
    def text(self) -> str:
        """Get the text of the result page."""
        return self.response.text

    @property
    def content(self) -> bytes:
        """Get the binary content of the result page."""
        return self.response.content

    @property
    def headers(self) -> CaseInsensitiveDict:
        """Get the headers of the result page."""
        return self.response.headers

    @property
    def title(self) -> Optional[str]:
        """Get the title of the result page."""
        try:
            return self.xpath('//title[1]/text()')[0]
        except IndexError:
            return None

    @property
    def links(self) -> List[Link]:
        """Get all links in the result page."""
        return [Link(a.text_content(), a.get('href'))
                for a in self.xpath('//a[@href]')]

    def find_link(self, pattern: str) -> Optional[Link]:
        """Find a link with a given pattern on the result page."""
        regex = re.compile(pattern)
        for link in self.links:
            if regex.search(link.text) or regex.search(link.url):
                return link
        return None

    def form(self, name_or_num: Union[str, int] = 1) -> Optional[FormElement]:
        """Get the form with the given name or number on the result page.

        Returns None if no such form can be found on the result page.
        """
        forms = self.forms

        if isinstance(name_or_num, str):

            # first, try ID
            for form in forms:
                form_id = form.get('id')
                if form_id and form_id == name_or_num:
                    return form

            # next, try regex with name
            regex = re.compile(name_or_num)
            for form in forms:
                name = form.get('name')
                if name and regex.search(name):
                    return form

        # last, try number
        try:
            num = int(name_or_num) - 1
            if not 0 <= num < len(forms):
                raise IndexError
        except (ValueError, IndexError):
            return None
        else:
            return forms[num]

    def _fix_forms(self) -> None:
        """Fix forms on the page for use with twill."""
        # put all stray fields into a form
        orphans = self.xpath('//input[not(ancestor::form)]')
        if orphans:
            form_parts = [b'<form>'] + [
                tree_to_html(orphan) for orphan in orphans] + [b'</form>']
            self.forms = html_to_tree(b''.join(form_parts)).forms
            self.forms.extend(self.tree.forms)
        else:
            self.forms = self.tree.forms
        # convert all submit button elements to input elements, since
        # otherwise lxml will not recognize them as form input fields
        for form in self.forms:
            for button in form.xpath("//button[@type='submit']"):
                button.tag = 'input'


def trunc(s: Optional[str], length: int) -> str:
    """Truncate a string to a given length.

    The string is truncated by cutting off the last (length-4) characters
    and replacing them with ' ...'
    """
    if s and len(s) > length:
        return s[:length - 4] + ' ...'
    return s or ''


def print_form(form: FormElement, n: int) -> None:
    """Pretty-print the given form, with the assigned number."""
    info = log.info
    name = form.get('name')
    info('\nForm name=%s (#%d)', name, n) if name else info('\nForm #%d', n)

    if form.inputs is not None:
        info('## __Name__________________'
             ' __Type___ __ID________ __Value__________________')

        for n, field in enumerate(form.inputs, 1):
            value = field.value
            value_options = getattr(field, 'value_options', None)
            if value_options:
                items = ', '.join(
                    f"'{getattr(opt, 'name', opt)}'"
                    for opt in value_options)
                value_displayed = f'{value} of {items}'
            else:
                value_displayed = f'{value}'
            field_name = field.name
            field_type = getattr(field, 'type', 'select')
            field_id = field.get('id')
            strings = (
                f'{n:2}',
                f'{trunc(field_name, 24):24}',
                f'{trunc(field_type, 9):9}',
                f'{trunc(field_id, 12):12}',
                trunc(value_displayed, 40))
            info(' '.join(strings))
    info('')


def make_boolean(value: Any) -> bool:
    """Convert the input value into a boolean."""
    value = str(value).lower().strip()

    # true/false
    if value in ('true', 'false'):
        return value == 'true'

    # 0/nonzero
    try:
        ival = int(value)
    except ValueError:
        pass
    else:
        return bool(ival)

    # +/-
    if value in ('+', '-'):
        return value == '+'

    # on/off
    if value in ('on', 'off'):
        return value == 'on'

    raise TwillException(f"unable to convert '{value}' into true/false")


def make_int(value: Any) -> int:
    """Convert the input value into an int."""
    try:
        ival = int(value)
    except Exception:
        pass
    else:
        return ival

    raise TwillException(f"unable to convert '{value}' into an int")


def set_form_control_value(control: FieldElement, value: str) -> None:
    """Set the given control to the given value

    The controls can be checkboxes, select elements etc.
    """
    if isinstance(control, InputElement):
        if control.checkable:
            try:
                boolean_value = make_boolean(value)
            except TwillException:
                # if there's more than one checkbox,
                # it should be a CheckboxGroup, see below.
                pass
            else:
                control.checked = boolean_value
        elif control.type not in ('submit', 'image'):
            control.value = value

    elif isinstance(control, (TextareaElement, RadioGroup)):
        control.value = value

    elif isinstance(control, CheckboxGroup):
        if value.startswith('-'):
            value = value[1:]
            try:
                control.value.remove(value)
            except KeyError:
                pass
        else:
            if value.startswith('+'):
                value = value[1:]
            control.value.add(value)

    elif isinstance(control, SelectElement):
        # for ListControls we need to find the right *value*,
        # and figure out if we want to *select* or *deselect*
        if value.startswith('-'):
            add = False
            value = value[1:]
        else:
            add = True
            if value.startswith('+'):
                value = value[1:]

        # now, select the value.
        option_values = [val.strip() for val in control.value_options]
        options = control.getchildren()  # type: ignore
        option_names = [(c.text or '').strip() for c in options]
        for name, opt in zip(option_names, option_values):
            if value not in (name, opt):
                continue
            if isinstance(control.value, MultipleSelectOptions):
                if add:
                    control.value.add(opt)
                elif opt in control.value:
                    control.value.remove(opt)
            else:
                control.value = opt if add else ""
            break
        else:
            raise TwillException('Attempt to set an invalid value')

    else:
        raise TwillException('Attempt to set value on invalid control')


def _all_the_same_submit(matches: Sequence[FieldElement]) -> bool:
    """Check if a list of controls all belong to the same control.

    For use with checkboxes, hidden, and submit buttons.
    """
    name = value = None
    for match in matches:
        if not isinstance(match, InputElement):
            return False
        if match.type not in ('submit', 'hidden'):
            return False
        if name is None:
            name = match.name
            value = match.value
        elif match.name != name or match.value != value:
            return False
    return True


def _all_the_same_checkbox(matches: Sequence[FieldElement]) -> bool:
    """Check if a list of controls all belong to the same checkbox.

    Hidden controls can combine with checkboxes, to allow form
    processors to ensure a False value is returned even if user
    does not check the checkbox. Without the hidden control, no
    value would be returned.
    """
    name = None
    for match in matches:
        if not isinstance(match, InputElement):
            return False
        if match.type not in ('checkbox', 'hidden'):
            return False
        if name is None:
            name = match.name
        else:
            if match.name != name:
                return False
    return True


def unique_match(matches: Sequence[FieldElement]) -> bool:
    """Check whether a match is unique"""
    return (len(matches) == 1 or
            _all_the_same_checkbox(matches) or _all_the_same_submit(matches))


def run_tidy(html: str) -> Tuple[Optional[str], Optional[str]]:
    """Run HTML Tidy on the given HTML string.

    Return a 2-tuple (output, errors).  (None, None) will be returned if
    PyTidyLib (or the required shared library for tidy) isn't installed.
    """
    from .commands import options
    require_tidy = options.get('require_tidy')

    if not tidylib:
        if require_tidy:
            raise TwillException(
                'Option require_tidy is set, but PyTidyLib is not installed')
        return None, None

    opts = {key[5:].replace('_', '-'): value
            for key, value in options.items() if key.startswith('tidy_')}
    clean_html, errors = tidylib.tidy_document(html, opts)
    return clean_html, errors


def get_equiv_refresh_interval() -> Optional[int]:
    """Get the smallest interval for which the browser should follow redirects.

    Redirection happens if the given interval is smaller than this.
    """
    from .commands import options
    return options.get('equiv_refresh_interval')


def is_hidden_filename(filename: str) -> bool:
    """Check if this is a hidden file (starting with a dot)."""
    return filename not in (
        '.', '..') and os.path.basename(filename).startswith('.')


def is_twill_filename(filename: str) -> bool:
    """Check if the given filename has the twill file extension."""
    return filename.endswith(twill_ext) and not is_hidden_filename(filename)


def make_twill_filename(name: str) -> str:
    """Add the twill extension to the name of a script if necessary."""
    if name not in ('.', '..'):
        twill_name, ext = os.path.splitext(name)
        if not ext:
            twill_name += twill_ext
            if os.path.exists(twill_name):
                name = twill_name
    return name


def gather_filenames(args: Sequence[str]) -> List[str]:
    """Collect script files from within directories."""
    names: List[str] = []
    for arg in args:
        name = make_twill_filename(arg)
        if os.path.isdir(name):
            for dir_path, dir_names, filenames in os.walk(arg):
                dir_names[:] = [
                    d for d in dir_names if not is_hidden_filename(d)]
                for filename in filenames:
                    if not is_twill_filename(filename):
                        continue
                    filename = os.path.join(dir_path, filename)
                    names.append(filename)
        else:
            names.append(name)
    return names



"""
Map of various User-Agent string shortcuts that can be used for testing.
"""

from typing import Dict

# noinspection HttpUrlsUsage
agents: Dict[str, str] = dict(
    # Desktop
    chrome_40='Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36'
              ' (KHTML, like Gecko) Chrome/40.0.2214.93 Safari/537.36',
    chrome_107='Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36'
               ' (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36',
    edge_12='Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36'
            ' (KHTML, like Gecko) Chrome/42.0.2311.135'
            ' Safari/537.36 Edge/12.246',
    edge_107='Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36'
             ' (KHTML, like Gecko) Chrome/107.0.0.0'
             ' Safari/537.36 Edg/107.0.1418.26',
    firefox_40='Mozilla/5.0 (Windows NT 6.1; WOW64; rv:40.0)'
               ' Gecko/20100101 Firefox/40.1',
    firefox_106='Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:106.0)'
                ' Gecko/20100101 Firefox/106.0',
    ie_3='Mozilla/2.0 (compatible; MSIE 3.0; Windows 3.1)',
    ie_4='Mozilla/4.0 (compatible; MSIE 4.0; Windows NT 5.0)',
    ie_5='Mozilla/4.0 (compatible; MSIE 5.0; Windows NT 5.0)',
    ie_6='Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)',
    ie_7='Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1)',
    ie_8='Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.0; Trident/4.0)',
    ie_9='Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0)',
    ie_10='Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/6.0)',
    ie_11='Mozilla/5.0 (Windows NT 6.3; Trident/7.0; rv:11.0) like Gecko',
    ie_mobile_9='Mozilla/5.0 (compatible; MSIE 9.0; Windows Phone OS 7.5;'
                ' Trident/5.0; IEMobile/9.0)',
    opera_7='Opera/7.0 (Windows NT 5.1; U) [en]',
    opera_8='Opera/8.00 (Windows NT 5.1; U; en)',
    opera_9='Opera/9.00 (Windows NT 5.2; U; en)',
    opera_10='Opera/9.80 (Windows NT 6.1; U; en) Presto/2.2.15 Version/10.00',
    opera_11='Opera/9.80 (Windows NT 6.1; U; en) Presto/2.7.62 Version/11.00',
    opera_12='Opera/12.0 (Windows NT 5.1; U; en)'
             ' Presto/22.9.168 Version/12.00',
    opera_mini_7='Opera/9.80 (Android; Opera Mini/7.0.29952/28.2075; en)'
                 ' Presto/2.8.119 Version/11.10',
    opera_mini_9='Opera/9.80 (J2ME/MIDP; Opera Mini/9 (Compatible; MSIE:9.0;'
                 ' iPhone; BlackBerry9700; AppleWebKit/24.746; en)'
                 ' Presto/2.5.25 Version/10.54',
    konqueror_3='Mozilla/5.0 (compatible; Konqueror/3.0; Linux)',
    konqueror_4='Mozilla/5.0 (compatible; Konqueror/4.0; Linux)'
                ' KHTML/4.0.3 (like Gecko)',
    lynx_2_8='Lynx/2.8.7rel.2 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/1.0.0a',
    w3m_0_5='w3m/0.5.2 (Linux i686; it; Debian-3.0.6-3)',
    netscape_3='Mozilla/3.0 (X11; I; AIX 2)',
    netscape_4='Mozilla/4.0 (compatible; Mozilla/5.0 ; Linux i686)',
    netscape_4_5='Mozilla/4.5 [en] (X11; I; SunOS 5.6 sun4u)',
    netscape_7='Mozilla/5.0 (X11; U; SunOS sun4u; en-US; rv:1.0.1)'
               ' Gecko/20020921 Netscape/7.0',
    netscape_9='Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.1.8pre)'
               ' Gecko/20071015 Firefox/2.0.0.7 Navigator/9.0',
    palemoon_25='Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:25.6)'
                ' Gecko/20150723 Firefox/31.9 PaleMoon/25.6.0',
    safari_1='Mozilla/5.0 (Macintosh; PPC Mac OS X; en)'
             ' AppleWebKit/85.7 (KHTML, like Gecko) Safari/85.6',
    safari_2='Mozilla/5.0 (Macintosh; PPC Mac OS; en)'
             ' AppleWebKit/412 (KHTML, like Gecko) Safari/412',
    safari_3='Mozilla/5.0 (Macintosh; Intel Mac OS X; en)'
             ' AppleWebKit/522.7 (KHTML, like Gecko) Version/3.0 Safari/522.7',
    safari_4='Mozilla/5.0 (Macintosh; U; PPC Mac OS X 10_5_6; en)'
             ' AppleWebKit/530.9+ (KHTML, like Gecko)'
             'Version/4.0 Safari/528.16',
    safari_5='Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_3; en)'
             ' AppleWebKit/534.1+ (KHTML, like Gecko)'
             ' Version/5.0 Safari/533.16',
    safari_6='Mozilla/5.0 (iPad; CPU OS 6_0 like Mac OS X) AppleWebKit/536'
             '.26 (KHTML, like Gecko)'
             ' Version/6.0 Mobile/10A5355d Safari/8536.25',
    safari_7='Mozilla/5.0 (iPad; CPU OS 7_1_2 like Mac OS X) AppleWebKit/537'
             '.51.2 (KHTML, like Gecko)'
             ' Version/7.0 Mobile/11D257 Safari/9537.53',
    safari_605='Mozilla/5.0 (Macintosh; Intel Mac OS X 13_0)'
               ' AppleWebKit/605.1.15 (KHTML, like Gecko)'
               ' Version/16.1 Safari/605.1.15',
    vivaldi_5='Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36'
              ' (KHTML, like Gecko) Chrome/107.0.0.0'
              ' Safari/537.36 Vivaldi/5.4.2753.51',
    # Android phones
    galaxy_s7='Mozilla/5.0 (Linux; Android 7.0; SM-G930VC Build/NRD90M; wv)'
              ' AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0'
              ' Chrome/58.0.3029.83 Mobile Safari/537.36',
    galaxy_s10='Mozilla/5.0 (Linux; Android 9; SM-G973U Build/PPR1.180610.011)'
               ' AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100'
               ' Mobile Safari/537.36',
    galaxy_s20='Mozilla/5.0 (Linux; Android 10;'
               ' SM-G980F Build/QP1A.190711.020; wv) AppleWebKit/537.36'
               ' (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.96'
               ' Mobile Safari/537.36',
    galaxy_s22='Mozilla/5.0 (Linux; Android 12;'
               ' SM-S906N Build/QP1A.190711.020; wv) AppleWebKit/537.36'
               ' (KHTML, like Gecko) Version/4.0 Chrome/80.0.3987.119'
               ' Mobile Safari/537.36',
    google_pixel='Mozilla/5.0 (Linux; Android 7.1.1; Google Pixel'
                 ' Build/NMF26F; wv) AppleWebKit/537.36 (KHTML, like Gecko)'
                 ' Version/4.0 Chrome/54.0.2840.85 Mobile Safari/537.36',
    google_pixel4='Mozilla/5.0 (Linux; Android 10; Google Pixel 4'
                  ' Build/QD1A.190821.014.C2; wv) AppleWebKit/537.36'
                  ' (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.108'
                  ' Mobile Safari/537.36',
    google_pixel_6='Mozilla/5.0 (Linux; Android 12; Pixel 6'
                   ' Build/SD1A.210817.023; wv) AppleWebKit/537.36'
                   ' (KHTML, like Gecko) Version/4.0 Chrome/94.0.4606.71'
                   ' Mobile Safari/537.36',
    nexus_6p='Mozilla/5.0 (Linux; Android 6.0.1; Nexus 6P Build/MMB29P)'
             ' AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.83'
             ' Mobile Safari/537.36',
    sony_xperia_1='Mozilla/5.0 (Linux; Android 9;'
                  ' J8110 Build/55.0.A.0.552; wv) AppleWebKit/537.36'
                  ' (KHTML, like Gecko) Version/4.0 Chrome/71.0.3578.99'
                  ' Mobile Safari/537.36',
    htc_one_x10='Mozilla/5.0 (Linux; Android 6.0; HTC One'
                ' X10 Build/MRA58K; wv) AppleWebKit/537.36'
                ' (KHTML, like Gecko) Version/4.0 Chrome/61.0.3163.98'
                ' Mobile Safari/537.36',
    # iPhones
    iphone_6='Mozilla/5.0 (Apple-iPhone7C2/1202.466; U; CPU like Mac OS X; en)'
             ' AppleWebKit/420+ (KHTML, like Gecko) Version/3.0'
             ' Mobile/1A543 Safari/419.3',
    iphone_7='Mozilla/5.0 (iPhone9,3; U; CPU iPhone OS 10_0_1 like Mac OS X)'
             ' AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0'
             ' Mobile/14A403 Safari/602.1',
    iphone_8='Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X)'
             ' AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0'
             ' Mobile/15A5341f Safari/604.1',
    iphone_x='Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X)'
             ' AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0'
             ' Mobile/15A372 Safari/604.1',
    iphone_11='Mozilla/5.0 (iPhone12,1; U; CPU iPhone OS 13_0 like Mac OS X)'
              ' AppleWebKit/602.1.50 (KHTML, like Gecko)'
              ' Version/10.0 Mobile/15E148 Safari/602.1',
    iphone_12='Mozilla/5.0 (iPhone13,2; U; CPU iPhone OS 14_0 like Mac OS X)'
              ' AppleWebKit/602.1.50 (KHTML, like Gecko)'
              ' Version/10.0 Mobile/15E148 Safari/602.1',
    iphone_13_pro_max='Mozilla/5.0 (iPhone14,3; U; CPU iPhone OS 15_0'
                      ' like Mac OS X) AppleWebKit/602.1.50'
                      ' (KHTML, like Gecko) Version/10.0'
                      ' Mobile/19A346 Safari/602.1',
    iphone_se_3='Mozilla/5.0 (iPhone14,6; U; CPU iPhone OS 15_4'
                ' like Mac OS X) AppleWebKit/602.1.50'
                ' (KHTML, like Gecko) Version/10.0 Mobile/19E241 Safari/602.1',
    # MS Windows phones
    ms_lumia_650='Mozilla/5.0 (Windows Phone 10.0; Android 6.0.1; Microsoft;'
                 ' RM-1152) AppleWebKit/537.36 (KHTML, like Gecko)'
                 ' Chrome/52.0.2743.116 Mobile Safari/537.36 Edge/15.15254',
    ms_lumia_950='Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Microsoft;'
                 ' Lumia 950) AppleWebKit/537.36 (KHTML, like Gecko)'
                 ' Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/13.1058',
    # Tablets
    galaxy_tab_s8='Mozilla/5.0 (Linux; Android 12;'
                  ' SM-X906C Build/QP1A.190711.020; wv) AppleWebKit/537.36'
                  ' (KHTML, like Gecko) Version/4.0 Chrome/80.0.3987.119'
                  ' Mobile Safari/537.36',
    lenovo_yoga_tab_11='Mozilla/5.0 (Linux; Android 11; Lenovo YT-J706X)'
                       ' AppleWebKit/537.36 (KHTML, like Gecko)'
                       ' Chrome/96.0.4664.45 Safari/537.36',
    sony_xperia_tab_z4='Mozilla/5.0 (Linux; Android 6.0.1;'
                       ' SGP771 Build/32.2.A.0.253; wv) AppleWebKit/537.36'
                       ' (KHTML, like Gecko) Version/4.0'
                       ' Chrome/52.0.2743.98 Safari/537.36',
    galaxy_tab_s3='Mozilla/5.0 (Linux; Android 7.0; SM-T827R4 Build/NRD90M)'
                  ' AppleWebKit/537.36 (KHTML, like Gecko)'
                  ' Chrome/60.0.3112.116 Safari/537.36',
    amazon_fire_hdx_7='Mozilla/5.0 (Linux; Android 4.4.3; KFTHWI Build/KTU84M)'
                      ' AppleWebKit/537.36 (KHTML, like Gecko) Silk/47.1.79'
                      ' like Chrome/47.0.2526.80 Safari/537.36',
    lg_g_pad_7='Mozilla/5.0 (Linux; Android 5.0.2; LG-V410/V41020c'
               ' Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko)'
               ' Version/4.0 Chrome/34.0.1847.118 Safari/537.36',
    # E-Readers
    kindle_4='Mozilla/5.0 (X11; U; Linux armv7l like Android; en-us)'
             ' AppleWebKit/531.2+ (KHTML, like Gecko) Version/5.0'
             ' Safari/533.2+ Kindle/3.0+',
    kindle_3='Mozilla/5.0 (Linux; U; en-US) AppleWebKit/528.5+'
             ' (KHTML, like Gecko, Safari/528.5+) Version/4.0 Kindle/3.0'
             ' (screen 600x800; rotate)',
    # Set tops
    chromecast='Mozilla/5.0 (CrKey armv7l 1.5.16041) AppleWebKit/537.36'
               ' (KHTML, like Gecko) Chrome/31.0.1650.0 Safari/537.36',
    amazon_4k_fire_tv='Mozilla/5.0 (Linux; Android 5.1; AFTS Build/LMY47O)'
                      ' AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0'
                      ' Chrome/41.99900.2250.0242 Safari/537.36',
    nexus_player='Dalvik/2.1.0 (Linux; U; Android 6.0.1;'
                 ' Nexus Player Build/MMB29T)',
    apple_tv_6='AppleTV11,1/11.1',
    apple_tv_5='AppleTV6,2/11.1',
    apple_tv_4='AppleTV5,3/9.1.1',
    # Game consoles
    playstation_5='Mozilla/5.0 (PlayStation; PlayStation 5/2.26)'
                  ' AppleWebKit/605.1.15 (KHTML, like Gecko)'
                  ' Version/13.0 Safari/605.1.15',
    playstation_4='Mozilla/5.0 (PlayStation 4 3.11) AppleWebKit/537.73'
                  ' (KHTML, like Gecko)',
    xbox_x='Mozilla/5.0 (Windows NT 10.0; Win64; x64; Xbox; Xbox Series X)'
           ' AppleWebKit/537.36 (KHTML, like Gecko)'
           ' Chrome/48.0.2564.82 Safari/537.36 Edge/20.02',
    xbox_one='Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Xbox; Xbox One)'
             ' AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0'
             ' Mobile Safari/537.36 Edge/13.10586',
    nintendo_switch='Mozilla/5.0 (Nintendo Switch; WifiWebAuthApplet)'
                    ' AppleWebKit/601.6 (KHTML, like Gecko) NF/4.0.0.5.10'
                    ' NintendoBrowser/5.1.0.13343',
    # Bots
    google_bot_2='Mozilla/5.0 (compatible; Googlebot/2.1;'
                 ' +http://www.google.com/bot.html)',
    bing_bot_2='Mozilla/5.0 (compatible; bingbot/2.0;'
               ' +http://www.bing.com/bingbot.htm)',
    yahoo_bot='Mozilla/5.0 (compatible; Yahoo! Slurp;'
              ' http://help.yahoo.com/help/us/ysearch/slurp)',
)



"""
An extension function to iterate over a list of comma-separated values.

Function 'csv_iterate' reads a file containing one or more rows of
comma-separated columns, assigns them to col1...colN, and, for each row,
executes the given twill script.
"""

import csv

from twill import execute_file, log, namespaces

__all__ = ['csv_iterate']


def csv_iterate(file_name, script_name):
    """>> csv_iterate <csv_file> <script>

    For each line in <csv_file>, read in a list of comma-separated values,
    put them in $col1...$colN, and execute <script>.
    """
    global_dict, local_dict = namespaces.get_twill_glocals()

    reader = csv.reader(open(file_name))
    for i, row in enumerate(reader, 1):
        log.debug('csv_iterate: on row %d of %s', i, file_name)
        for j, col in enumerate(row, 1):
            global_dict[f"col{j}"] = col

        execute_file(script_name, no_reset=True)



"""
Implementation of all the individual 'twill' commands available
through twill-sh.
"""

import getpass
import re
import time
import sys
from typing import Any, Dict, Optional

from os.path import sep

from . import log, set_output, set_err_out, utils
from .agents import agents
from .browser import browser
from .errors import TwillException, TwillAssertionError
from .namespaces import get_twill_glocals

# noinspection SpellCheckingInspection
__all__ = [
    'add_auth', 'add_cleanup', 'add_extra_header', 'agent',
    'back', 'browser',
    'clear_cookies', 'clear_extra_headers',
    'code', 'config',
    'debug', 'echo', 'exit', 'extend_with',
    'find', 'follow',
    'form_action', 'formaction', 'fa',
    'form_clear', 'formclear', 'form_file', 'formfile',
    'form_value', 'formvalue', 'fv',
    'get_input', 'getinput', 'get_password', 'getpassword',
    'go', 'info', 'load_cookies', 'not_find', 'notfind', 'options',
    'redirect_error', 'redirect_output',
    'reload', 'reset_browser', 'reset_error', 'reset_output',
    'run', 'run_file', 'runfile', 'rf',
    'save_cookies', 'save_html',
    'setglobal', 'set_global', 'setlocal', 'set_local',
    'show', 'showcookies', 'show_cookies', 'show_extra_headers',
    'showforms', 'show_forms', 'showhistory', 'show_history',
    'showhtml', 'show_html', 'showlinks', 'show_links',
    'sleep', 'submit',
    'tidy_ok', 'title', 'url']


def reset_browser():
    """>> reset_browser

    Reset the browser completely.
    """
    browser.reset()
    options.clear()
    options.update(default_options)


# noinspection PyShadowingBuiltins
def exit(code: str = '0') -> None:
    """twill command: exit [<code>]

    Exit twill, with the given exit code (defaults to 0, "no error").
    """
    raise SystemExit(int(code))


def go(url: str) -> None:
    """>> go <url>

    Visit the URL given.
    """
    browser.go(url)


def reload() -> None:
    """>> reload

    Reload the current URL.
    """
    browser.reload()


def code(should_be: str) -> None:
    """>> code <int>

    Check to make sure the response code for the last page is as given.
    """
    if browser.code != int(should_be):
        raise TwillAssertionError(f"code is {browser.code} != {should_be}")


def tidy_ok() -> None:
    """>> tidy_ok

    Assert that 'tidy' does not produce any warnings or errors when run on
    the current page.

    If 'tidy' cannot be run, will fail silently (unless 'require_tidy' option
    is true; see 'config' command).
    """
    page = browser.html
    if page is None:
        raise TwillAssertionError("not viewing HTML!")

    clean_page, errors = utils.run_tidy(page)
    if clean_page is None:  # tidy doesn't exist...
        if options.get('require_tidy'):
            raise TwillAssertionError("cannot run 'tidy'")
    elif errors:
        raise TwillAssertionError(f"tidy errors:\n====\n{errors}\n====\n")


def url(should_be: str) -> str:
    """>> url <pattern>

    Check to make sure that the current URL matches the regex pattern.
    The local variable __match__ is set to the matching part of the URL.
    """
    regex = re.compile(should_be)
    current_url = browser.url

    if current_url is None:
        current_url = ''
        m = None
    else:
        m = regex.search(current_url)

    if not m:
        raise TwillAssertionError(
            f"current url is '{current_url}';\n"
            f"does not match '{should_be}'\n")

    match_str = m.group(1 if m.groups() else 0)
    global_dict, local_dict = get_twill_glocals()
    local_dict['__match__'] = match_str
    return match_str


def follow(what: str) -> str:
    """>> follow <pattern>

    Find the first link on the page matching the given regex pattern and
    then visit it.
    """
    link = browser.find_link(what)
    if link:
        browser.follow_link(link)
        return browser.url

    raise TwillAssertionError(f"no links match to '{what}'")


_find_flags = dict(i=re.IGNORECASE, m=re.MULTILINE, s=re.DOTALL)


def _parse_find_flags(flags: str) -> int:
    """Helper function to parse the find flags."""
    re_flags = 0
    for char in flags:
        try:
            re_flags |= _find_flags[char]
        except IndexError:
            raise TwillAssertionError(f"unknown 'find' flag {char!r}")
    return re_flags


def find(what: str, flags='') -> str:
    """>> find <pattern> [<flags>]

    Succeed if the regular expression pattern can be found on the page.
    Sets the local variable __match__ to the matching text.

    Flags is a string consisting of the following characters:

    * i: ignore case
    * m: multi-line
    * s: dot matches all
    * x: use XPath expressions instead of regular expression

    For explanations of regular expressions, please see the Python re module
    documentation.
    """
    page = browser.html
    local_dict = get_twill_glocals()[1]
    if 'x' in flags:
        elements = browser.xpath(what)
        if not elements:
            raise TwillAssertionError(f"no element to path '{what}'")
        match_str = elements[0].text or ''
    else:
        match = re.search(what, page, flags=_parse_find_flags(flags))
        if not match:
            raise TwillAssertionError(f"no match to '{what}'")
        match_str = match.group(1 if match.groups() else 0)
    local_dict['__match__'] = match_str
    return match_str


def not_find(what: str, flags='') -> None:
    """>> not_find <pattern> [<flags>]

    Fail if the regular expression pattern can be found on the page.
    """
    try:
        find(what, flags)
    except TwillAssertionError:
        pass
    else:
        raise TwillAssertionError(f"match to '{what}'")


# noinspection SpellCheckingInspection
notfind = not_find  # backward compatibility and convenience


def back() -> None:
    """>> back

    Return to the previous page.
    """
    browser.back()


def show(what: Optional[str] = None) -> None:
    """>> show [<objects>]

    Show the specified objects (html, cookies, forms, links, history).
    """
    if not what:
        what = 'html'
    command = None
    if what.isalpha():
        command_name = f'show_{what}'
        if command_name in __all__:
            command = globals().get(command_name)
    if not command:
        raise TwillException(f'Cannot show "{what}".')
    command()


def show_html() -> None:
    """>> show_html

    Show the HTML for the current page or show the specified objects
    (which can be cookies, forms, history or links).

    Note: Use browser.html to get the HTML programmatically.
    """
    html = browser.html.strip()
    log.info('')
    log.info(html)
    log.info('')


# noinspection SpellCheckingInspection
showhtml = show_html  # backward compatibility and consistency


def echo(*strs: str) -> None:
    """>> echo <list> <of> <strings>

    Echo the arguments to the screen.
    """
    log.info(' '.join(map(str, strs)))


def save_html(filename: Optional[str] = None) -> None:
    """>> save_html [<filename>]

    Save the HTML for the current page into <filename>.
    If no filename given, construct the filename from the URL.
    """
    html = browser.html
    if html is None:
        log.warning("No page to save.")
        return

    if filename is None:
        url = browser.url
        url = url.split('?', 1)[0]
        filename = url.rsplit('/', 1)[-1]
        if not filename:
            filename = 'index.html'
        log.info("Using filename '%s'.", filename)

    encoding = browser.encoding or 'utf-8'
    try:
        with open(filename, 'w', encoding=encoding) as f:
            f.write(html)
    except UnicodeEncodeError:
        if encoding == 'utf-8':
            raise
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(html)


def sleep(interval: str = "1") -> None:
    """>> sleep [<interval>]

    Sleep for the specified amount of time.
    If no interval is given, sleep for 1 second.
    """
    time.sleep(float(interval))


def agent(what: str) -> None:
    """>> agent <agent>

    Set the agent string (identifying the browser brand).

    Some convenient shortcuts:
    chrome_107, firefox_106, safari_605, edge_107, ie_11.

    See twill.agents for a list of all available shortcuts.
    """
    what = what.strip()
    agent = agents.get(what, what)
    browser.agent_string = agent


def submit(submit_button: Optional[str] = None,
           form_name: Optional[str] = None) -> None:
    """>> submit [<submit_button> [<form_name>]]

    Submit the current form (the one last clicked on) by clicking on the
    given submission button.  If no 'submit_button' is given, submit the
    current form by using the last clicked submit button.

    The form to submit is the last form clicked on with a 'form_value' command
    unless explicitly specified given the

    The button used to submit is chosen based on 'submit_button'.
    If 'submit_button' is given, it's matched against buttons using
    the same rules that 'form_value' uses.  If 'button_name' is not given,
    this function uses the last submit button clicked on by 'form_value'.
    If none can be found, it submits the form with no submit button clicked.
    """
    browser.submit(submit_button, form_name)


def show_forms() -> None:
    """>> show_forms

    Show all the forms on the current page.

    Note: Use browser.forms to get the forms programmatically.
    """
    browser.show_forms()


# noinspection SpellCheckingInspection
showforms = show_forms  # backward compatibility and convenience


def show_links() -> None:
    """>> show_links

    Show all the links on the current page.

    Note: Use browser.links to get the links programmatically.
    """
    browser.show_links()


# noinspection SpellCheckingInspection
showlinks = show_links  # backward compatibility and convenience


def show_history() -> None:
    """>> show_history

    Show the browser history (what URLs were visited).

    Note: Use browser.history to get the history programmatically.
    """
    browser.show_history()


# noinspection SpellCheckingInspection
showhistory = show_history  # backward compatibility and convenience


def form_clear(form_name: str) -> None:
    """>> form_clear <form_name>

    Run 'clear' on all the controls in this form.
    """
    form = browser.form(form_name)
    if form is None:
        raise TwillAssertionError("Form not found")
    for control in form.inputs:
        if not ('readonly' in control.attrib
                or 'disabled' in control.attrib
                or getattr(control, 'type', None)
                in ('submit', 'image', 'hidden')):
            del control.value
    browser.last_submit_button = None


# noinspection SpellCheckingInspection
formclear = form_clear  # backward compatibility and convenience


def form_value(form_name: str, field_name: str, value: str) -> None:
    """>> form_value <form_name> <field_name> <value>

    Set value of a form field.

    There are some ambiguities in the way 'form_value' deals with lists:
    'form_value' will *add* the given value to a list of multiple selection,
    for lists that allow it.

    Forms are matched against 'form_name' as follows:
      1. regex match to actual form name;
      2. if 'form_name' is an integer, it's tried as an index.

    Form controls are matched against 'field_name' as follows:
      1. unique exact match to control name;
      2. unique regex match to control name;
      3. if field_name is an integer, it's tried as an index;
      4. unique & exact match to submit-button values.

    'form_value' ignores read-only fields completely; if they're readonly,
    nothing is done, unless the config options ('config' command) are
    changed.

    'form_value' is available as 'fv' as well.
    """
    form = browser.form(form_name)
    if form is None:
        raise TwillAssertionError("Form not found")

    control = browser.form_field(form, field_name)

    browser.clicked(form, control)

    attrib = getattr(control, 'attrib', {})
    if 'readonly' in attrib:
        if options['readonly_controls_writeable']:
            log.info('Forcing read-only form field to writeable.')
            del attrib['readonly']
        else:
            log.info('Form field is read-only or ignorable; nothing done.')
            return

    if getattr(control, 'type', None) == 'file':
        raise TwillException(
            'form field is for file upload; use "form_file" instead')

    value = browser.decode(value)
    utils.set_form_control_value(control, value)


# noinspection SpellCheckingInspection
fv = formvalue = form_value  # backward compatibility and convenience


def form_action(form_name: str, action_url: str) -> None:
    """>> form_action <form_name> <action_url>

    Sets action parameter on form to action_url.

    'form_action' is available as 'fa' as well.
    """
    form = browser.form(form_name)
    if form is None:
        raise TwillAssertionError("Form not found")
    log.info("Setting action for form %s to %s.", form, action_url)
    form.action = action_url


# noinspection SpellCheckingInspection
fa = formaction = form_action  # backward compatibility and convenience


def form_file(form_name: str, field_name: str, filename: str,
              content_type: Optional[str] = None) -> None:
    """>> form_file <form_name> <field_name> <filename> [<content_type>]

    Upload a file via an "upload file" form field.
    """
    filename = filename.replace('/', sep)

    form = browser.form(form_name)
    if form is None:
        raise TwillAssertionError("Form not found")
    control = browser.form_field(form, field_name)

    if getattr(control, 'type', None) != 'file':
        raise TwillException('ERROR: field is not a file upload field!')

    browser.clicked(form, control)
    plain = content_type and content_type.startswith(('plain/', 'html/'))
    fp = open(filename, 'r' if plain else 'rb')
    browser.add_form_file(field_name, fp)

    log.info(
        'Added file "%s" to file upload field "%s".', filename, field_name)


# noinspection SpellCheckingInspection
formfile = form_file  # backward compatibility and convenience


def extend_with(module_name: str) -> None:
    """>> extend_with <module_name>

    Import contents of given module.
    """
    global_dict, local_dict = get_twill_glocals()

    exec(f"from {module_name} import *", global_dict)

    # now add the commands into the commands available for the shell,
    # and print out some nice stuff about what the extension module does.

    mod = sys.modules[module_name]

    from . import parse, shell

    fn_list = getattr(mod, '__all__', None)
    if fn_list is None:
        fn_list = [fn for fn in dir(mod) if callable(getattr(mod, fn))]

    for command in fn_list:
        fn = getattr(mod, command)
        shell.add_command(command, fn.__doc__)
        parse.command_list.append(command)

    info, debug = log.info, log.debug
    info("Imported extension module '%s'.", module_name)
    debug("(at %s)", mod.__file__)

    if shell.interactive:
        if mod.__doc__:
            info("\nDescription:\n\n%s\n", mod.__doc__.strip())
        else:
            if fn_list:
                info('New commands:\n')
                for name in fn_list:
                    info('\t%s', name)
                info('')


def get_input(prompt: str) -> str:
    """>> get_input <prompt>

    Get input, store it in '__input__'.
    """
    local_dict = get_twill_glocals()[1]

    inp = input(prompt)

    local_dict['__input__'] = inp
    return inp


# noinspection SpellCheckingInspection
getinput = get_input  # backward compatibility and convenience


def get_password(prompt: str) -> str:
    """>> get_password <prompt>

    Get a password ("invisible input"), store it in '__password__'.
    """
    local_dict = get_twill_glocals()[1]

    # we use sys.stdin here in order to get the same behaviour on Unix
    # as on other platforms and for better testability of this function
    inp = getpass.getpass(prompt, sys.stdin)

    local_dict['__password__'] = inp
    return inp


# noinspection SpellCheckingInspection
getpassword = get_password  # backward compatibility and convenience


def save_cookies(filename: str) -> None:
    """>> save_cookies <filename>

    Save all the current cookies to the given file.
    """
    browser.save_cookies(filename)


def load_cookies(filename: str) -> None:
    """>> load_cookies <filename>

    Clear the cookie jar and load cookies from the given file.
    """
    browser.load_cookies(filename)


def clear_cookies() -> None:
    """>> clear_cookies

    Clear the cookie jar.
    """
    browser.clear_cookies()


def show_cookies() -> None:
    """>> show_cookies

    Show all the cookies in the cookie jar.

    Note: Use browser.cookies to get the cookies programmatically.
    """
    browser.show_cookies()


# noinspection SpellCheckingInspection
showcookies = show_cookies  # backward compatibility and convenience


def add_auth(realm: str, uri: str, user: str, passwd: str) -> None:
    """>> add_auth <realm> <uri> <user> <passwd>

    Add HTTP Basic Authentication information for the given realm/uri.
    """
    if realm is not None:
        browser.add_creds((uri, realm), user, passwd)
        log.info(
            "Added auth info: realm '%s' / URI '%s' / user '%s'.",
            realm, uri, user)
    if realm is None or options['with_default_realm']:
        browser.add_creds(uri, user, passwd)
        if realm is None:
            log.info("Added auth info: URI '%s' / user '%s'.", uri, user)


def debug(what: str, level: str) -> None:
    """>> debug <what> <level>

    <what> can be:
       * http (any level >= 1), to display the HTTP transactions.
       * commands (any level >= 1), to display the commands being executed.
       * equiv-refresh (any level >= 1) to display HTTP-EQUIV refresh handling.
    """
    from . import parse

    try:
        num_level = int(level)
    except ValueError:
        num_level = 1 if utils.make_boolean(level) else 0

    log.info('DEBUG: Setting %s debugging to level %d.', what, num_level)

    if what == 'http':
        browser.debug_level = num_level
    elif what == 'equiv-refresh':
        browser.show_refresh = num_level > 0
    elif what == 'commands':
        parse.log_commands(num_level > 0)
    else:
        raise TwillException(f'Unknown debugging type: "{what}"')


def run(cmd: str) -> None:
    """>> run <command>

    <command> can be any valid Python command; 'exec' is used to run it.
    """
    # @CTB: use pyparsing to grok the command?  make sure that quoting works...

    # execute command.
    global_dict, local_dict = get_twill_glocals()

    # set __url__
    local_dict['__cmd__'] = cmd
    local_dict['__url__'] = browser.url

    exec(cmd, global_dict, local_dict)


def run_file(*args: str) -> None:
    """>> run_file <file1> [<file2> ...]

    Execute the given twill scripts or directories of twill scripts.

    'run_file' is available as 'rf' as well.
    """
    from . import parse

    filenames = utils.gather_filenames(args)
    for filename in filenames:
        parse.execute_file(filename, no_reset=True)


# noinspection SpellCheckingInspection
rf = runfile = run_file  # backward compatibility and convenience


def add_cleanup(*args: str) -> None:
    """>> add_cleanup <file1> [<file2> ...]

    Execute the given twill scripts after the current twill script.
    """

    local_dict = get_twill_glocals()[1]
    cleanups = local_dict.setdefault('__cleanups__', [])
    filenames = utils.gather_filenames(args)
    log.debug('Adding cleanup scripts: %s', ', '.join(filenames))
    cleanups.extend(reversed(filenames))


def set_global(name: str, value: str) -> None:
    """set_global <name> <value>

    Sets the variable <name> to the value <value> in the global namespace.
    """
    global_dict, local_dict = get_twill_glocals()
    global_dict[name] = value


# noinspection SpellCheckingInspection
setglobal = set_global  # backward compatibility and convenience


def set_local(name: str, value: str) -> None:
    """set_local <name> <value>

    Sets the variable <name> to the value <value> in the local namespace.
    """
    global_dict, local_dict = get_twill_glocals()
    local_dict[name] = value


# noinspection SpellCheckingInspection
setlocal = set_local  # backward compatibility and convenience


def title(what: str) -> str:
    """>> title <pattern>

    Succeed if the regular expression pattern is in the page title.
    """
    regex = re.compile(what)
    title = browser.title

    if title is None:
        log.info("The page has no title.")
    else:
        log.info("The title is '%s'.", title)

    m = regex.search(title) if title else None
    if m is None:
        raise TwillAssertionError(f"The title does not contain '{what}'.")

    if m.groups():
        match_str = m.group(1)
    else:
        match_str = m.group(0)

    global_dict, local_dict = get_twill_glocals()
    local_dict['__match__'] = match_str
    return match_str


def redirect_output(filename: str) -> None:
    """>> redirect_output <filename>

    Append all twill output to the given file.
    """
    fp = open(filename, 'a')
    set_output(fp)


def reset_output() -> None:
    """>> reset_output

    Reset twill output to go to the screen.
    """
    set_output(None)


def redirect_error(filename: str) -> None:
    """>> redirect_error <filename>

    Append all twill error output to the given file.
    """
    fp = open(filename, 'a')
    set_err_out(fp)


def reset_error() -> None:
    """>> reset_error

    Reset twill error output to go to the screen.
    """
    set_err_out(None)


def add_extra_header(header_key: str, header_value: str) -> None:
    """>> add_header <name> <value>

    Add an HTTP header to each HTTP request.  See 'show_extra_headers' and
    'clear_extra_headers'.
    """
    browser.headers[header_key] = header_value


def show_extra_headers() -> None:
    """>> show_extra_headers

    Show any extra headers being added to each HTTP request.
    """
    info = log.info
    headers = browser.headers
    if headers:
        info('\nThe following HTTP headers are added to each request:\n')
        for key, value in headers.items():
            info('\t"%s" = "%s"', key, value)
        info('')
    else:
        info('** no extra HTTP headers **')


def clear_extra_headers() -> None:
    """>> clear_extra_headers

    Remove all user-defined HTTP headers.  See 'add_extra_header' and
    'show_extra_headers'.
    """
    browser.reset_headers()


default_options: Dict[str, Any] = dict(
    equiv_refresh_interval=2,
    readonly_controls_writeable=False,
    require_tidy=False,
    with_default_realm=False)

options = default_options.copy()  # the global options dictionary


def config(key: Optional[str] = None, value: Any = None) -> None:
    """>> config [<key> [<int value>]]

    Configure/report various options.  If no <value> is given, report
    the current key value; if no <key> given, report current settings.

    Options starting with "tidy_" will be used to configure HTML tidy.

    So far:

     * 'equiv_refresh_interval', default 2 -- time limit for HTTP-EQUIV=REFRESH
     * 'readonly_controls_writeable', default False -- all controls writeable
     * 'require_tidy', default False -- *require* that tidy be installed
     * 'with_default_realm', default False -- use a default realm for HTTP AUTH
    """
    info = log.info
    if key is None:
        keys = sorted(options)
        info('\nCurrent configuration:\n')
        for k in keys:
            info('\t%s : %s', k, options[k])
        info('')
    else:
        v = options.get(key)
        if v is None and not key.startswith('tidy_'):
            log.error("no such configuration key '%s'", key)
            info("valid keys are: %s", ', '.join(sorted(options)))
            raise TwillException("no such configuration key: '%s'" % (key,))
        elif value is None:
            info('\nkey %s: value %s\n', key, v)
        else:
            if isinstance(v, bool):
                value = utils.make_boolean(value)
            elif isinstance(v, int):
                value = utils.make_int(value)
            options[key] = value


def info() -> None:
    """>> info

    Report information on current page.
    """
    current_url = browser.url
    if current_url is None:
        log.warning("We're not on a page!")
        return

    content_type = browser.response_headers['content-type']
    is_html = content_type and content_type.split(';', 1)[0] == 'text/html'
    code = browser.code

    info = log.info
    info('\tURL: %s', current_url)
    info('\tHTTP code: %s', code)
    info('\tContent type: %s%s', content_type, ' (HTML)' if is_html else '')
    if is_html:
        title = browser.title
        info('\tPage title: %s', title)
        forms = browser.forms
        if len(forms):
            info('\tThis page contains %d form(s)', len(forms))
    info('')



"""
Extension functions to discard all moderated messages in a SourceForge-based
mailman queue.

(Currently there is no way to do this without manually selecting 'discard'
for each and every message.)
"""

import re

from twill import browser, log, utils

__all__ = ['discard_all_messages', 'exit_if_empty']


def exit_if_empty():
    """>> exit_if_empty

    Exit the script currently running, if there are no deferred messages
    on the current page.
    """
    form = browser.form()
    if not form:
        log.error("No messages; exiting.")
        raise SystemExit


def discard_all_messages():
    """>> discard_all_messages

    Set all buttons to "discard".
    """
    _form_value_by_regex_setall('1', '^\\d+$', '3')


def _form_value_by_regex_setall(form_name, field_name, value):
    form = browser.form(form_name)
    if not form:
        log.error("no such form '%s'", form_name)
        return

    regex = re.compile(field_name)

    matches = [ctl for ctl in form.controls if regex.search(str(ctl.name))]

    if matches:
        log.info('-- matches %d', len(matches))

        n = 0
        for control in matches:
            browser.clicked(form, control)
            if not control.readonly:
                utils.set_form_control_value(control, value)
                n += 1

        log.info('set %d values total', n)



"""
Extension functions to check all of the links on a page.

Usage:

   check_links [ <pattern> ]

Make sure that all the HTTP links on the current page can be visited
successfully.  If 'pattern' is given, check only URLs that match that
regular expression.

If option 'check_links.only_collect_bad_links' is on, then all bad
links are silently collected across all calls to check_links.  The
function 'report_bad_links' can then be used to report all the links,
together with their referring pages.
"""

import re

from typing import Dict, List, Set

from twill import browser, commands, log, utils
from twill.errors import TwillAssertionError

__all__ = ['check_links', 'report_bad_links', 'good_urls', 'bad_urls']

# first, set up config options & persistent 'bad links' memory...
if commands.options.get('check_links.only_collection_bad_links') is None:
    commands.options['check_links.only_collect_bad_links'] = False

good_urls: Set[str] = set()
bad_urls: Dict[str, Set[str]] = dict()


def check_links(pattern=''):
    """>> check_links [<pattern>]

    Make sure that all the HTTP links on the current page can be visited
    with an HTTP response 200 (success).  If 'pattern' is given, interpret
    it as a regular expression that link URLs must contain in order to be
    tested, e.g.

        check_links https://.*\\.google\\.com

    would check only links to google URLs.  Note that because 'follow'
    is used to visit the pages, the referrer URL is properly set on the
    visit.
    """
    debug, info = log.debug, log.info

    debug('in check_links')

    # compile the regex
    regex = re.compile(pattern) if pattern else None

    # iterate over all links, collecting those that match
    #
    # note that in the case of duplicate URLs, only one of the
    # links is actually followed!

    collected_urls: Set[str] = set()

    links = browser.links
    if not links:
        debug("no links to check!?")
        return

    for link in links:
        url = link.url
        url = url.split('#', 1)[0]  # get rid of subpage pointers

        # noinspection HttpUrlsUsage
        if not url.startswith(('http://', 'https://')):
            debug("url '%s' is not an HTTP link; ignoring", url)
            continue

        if regex:
            if regex.search(url):
                collected_urls.add(url)
                debug("Gathered URL %s -- matched pattern", url)
            else:
                debug("URL %s doesn't match pattern", url)
        else:
            collected_urls.add(url)
            debug("Gathered URL %s.", url)

    # now, for each unique and unchecked URL, follow the link

    failed: List[str] = []
    for url in sorted(collected_urls):
        debug("Checking %s", url)
        if url in good_urls:
            debug('... already known as good')
        elif url in bad_urls:
            debug('... already collected as broken')
        else:
            try:
                browser.follow_link(url)
            except Exception:  # count as failure
                code = 404
            else:
                code = browser.code
                browser.back()
            if code == 200:
                debug('...success!')
                good_urls.add(url)
            else:
                debug('...failure!')
                failed.append(url)

    if commands.options['check_links.only_collect_bad_links']:
        for url in failed:
            referrers = bad_urls.setdefault(url, set())
            info('*** %s', browser.url)
            referrers.add(browser.url)
    elif failed:
        info('\nCould not follow %d links:\n', len(failed))
        for url in failed:
            info('* %s', url)
        raise TwillAssertionError("broken links on page")
    else:
        info('\nNo broken links were detected.\n')


def report_bad_links(fail_if_exist='true', flush_bad_links='true'):
    """>> report_bad_links [<fail-if-exist> [<flush-bad-links>]]

    Report all the links collected across check_links runs (collected
    if and only if the config option check_links.only_collect_bad_links
    is set).

    If <fail-if-exist> is false (true by default) then the command will
    fail after reporting any bad links.

    If <flush-bad-links> is false (true by default) then the list of
    bad links will be retained across the function call.
    """
    fail_if_exist = utils.make_boolean(fail_if_exist)
    flush_bad_links = utils.make_boolean(flush_bad_links)

    info = log.info
    if not bad_urls:
        info('\nNo bad links to report.\n')
        return

    info('\nCould not follow %d links', len(bad_urls))
    for url in sorted(bad_urls):
        referrers = sorted(bad_urls[url])
        info("\tlink '%s' (occurs on: %s)", url, ','.join(referrers))

    if flush_bad_links:
        bad_urls.clear()

    if fail_if_exist:
        raise TwillAssertionError("broken links encountered")



"""This module implements the TwillBrowser."""

import pickle
import re
from typing import (
    cast, Callable, Dict, IO, List, Optional, Sequence, Tuple, Union)
from urllib.parse import urljoin

from requests import Session
from requests.auth import HTTPBasicAuth
from requests.cookies import RequestsCookieJar
from requests.exceptions import InvalidSchema, ConnectionError
from requests.structures import CaseInsensitiveDict

from . import log, __version__
from .utils import (
    get_equiv_refresh_interval, html_to_tree, print_form, trunc, unique_match,
    CheckboxGroup, FieldElement, FormElement, HtmlElement,
    InputElement, Link, UrlWithRealm, RadioGroup, Response, ResultWrapper)
from .errors import TwillException

__all__ = ['browser']


def _disable_insecure_request_warnings() -> None:
    """Disable insecure request warnings."""
    try:
        from requests.packages import urllib3  # type: ignore
    except ImportError:
        import urllib3  # type: ignore
    # noinspection PyUnresolvedReferences
    insecure_request_warning = urllib3.exceptions.InsecureRequestWarning
    urllib3.disable_warnings(insecure_request_warning)


def _set_http_connection_debuglevel(level: int) -> None:
    """Set the debug level for the connection pool."""
    from http.client import HTTPConnection
    HTTPConnection.debuglevel = level


class TwillBrowser:
    """A simple, stateful browser"""

    user_agent = f'TwillBrowser/{__version__}'

    def __init__(self):
        self.result: Optional[ResultWrapper] = None
        self.last_submit_button: Optional[InputElement] = None
        self.first_error: Optional[str] = None

        # whether meta refresh will be displayed
        self.show_refresh = False

        # debug level to be used for the connection pool
        self._debug_level = 0

        # whether the SSL cert will be verified, or can be a ca bundle path
        self.verify = False

        # Session stores cookies
        self._session = Session()

        # A lxml FormElement, None until a form is selected
        # replaces self._browser.form from mechanize
        self._form: Optional[FormElement] = None
        self._form_files: Dict[str, IO] = {}

        # A dict of HTTPBasicAuth from requests, keyed off URL
        self._auth: Dict[UrlWithRealm, HTTPBasicAuth] = {}

        # callables to be called after each page load.
        self._post_load_hooks: List[Callable] = []

        self._history: List[ResultWrapper] = []

        # set default headers
        self.reset_headers()

    def _assert_result_for(self, what: str) -> ResultWrapper:
        if not self.result:
            raise TwillException(f"Cannot get {what} since there is no page.")
        return self.result

    @property
    def debug_level(self) -> int:
        return self._debug_level

    @debug_level.setter
    def debug_level(self, level: int) -> None:
        _set_http_connection_debuglevel(level)
        self._debug_level = level

    def reset(self):
        """Reset the browser"""
        self.__init__()

    @property
    def creds(self) -> Dict[UrlWithRealm, HTTPBasicAuth]:
        """Get the credentials for basic authentication."""
        return self._auth

    def add_creds(self, url: UrlWithRealm, user: str, password: str) -> None:
        """Set the credentials for basic authentication."""
        self._auth[url] = HTTPBasicAuth(user, password)

    def go(self, url: str) -> None:
        """Visit given URL."""
        try_urls: List[str] = []
        if '://' in url:
            try_urls.append(url)
        else:  # URL does not have a schema
            # if this is a relative URL, then assume that we want to tack it
            # onto the end of the current URL
            current_url = self.url
            if current_url:
                try_urls.append(urljoin(current_url, url))
            # if this is an absolute URL, it may be just missing the 'http://'
            # at the beginning, try fixing that (mimic browser behavior)
            if not url.startswith(('.', '/', '?')):
                # noinspection HttpUrlsUsage
                try_urls.append(f'http://{url}')
                try_urls.append(f'https://{url}')
        for try_url in try_urls:
            try:
                self._journey('open', try_url)
            except (IOError,
                    ConnectionError, InvalidSchema, UnicodeError) as error:
                log.info("cannot go to '%s': %s", try_url, error)
            else:
                break
        else:
            raise TwillException(f"cannot go to '{url}'")
        log.info('==> at %s', self.url)

    def reload(self) -> None:
        """Tell the browser to reload the current page."""
        self._journey('reload')
        log.info('==> reloaded')

    def back(self) -> None:
        """Return to previous page, if possible."""
        try:
            self._journey('back')
            log.info('==> back to %s', self.url)
        except TwillException:
            log.warning('==> back at empty page')

    @property
    def code(self) -> int:
        """Get the HTTP status code received for the current page."""
        return self._assert_result_for('status code').http_code

    @property
    def encoding(self) -> Optional[str]:
        """Get the encoding used by the server for the current page."""
        return None if self.result is None else self.result.encoding

    @property
    def html(self) -> str:
        """Get the HTML for the current page."""
        return self._assert_result_for('HTML').text

    @property
    def dump(self) -> bytes:
        """Get the binary content of the current page."""
        return self._assert_result_for('content dump').content

    @property
    def title(self) -> Optional[str]:
        return self._assert_result_for('title').title

    @property
    def url(self):
        """Get the URL of the current page."""
        return self.result.url if self.result else None

    def find_link(self, pattern: str) -> Optional[Link]:
        """Find the first link matching the given regular expression pattern.

        The pattern is searched in the URL and in the link text.
        """
        return self._assert_result_for('links').find_link(pattern)

    def follow_link(self, link: Union[str, Link]) -> None:
        """Follow the given link."""
        self._journey('follow_link', link)
        log.info('==> at %s', self.url)

    @property
    def headers(self) -> CaseInsensitiveDict:
        """Return the request headers currently used by the browser."""
        return self._session.headers

    def reset_headers(self):
        """Reset the request headers currently used by the browser."""
        self.headers.clear()
        self.headers.update({
            'Accept': 'text/html; */*',
            'User-Agent': self.user_agent})

    @property
    def response_headers(self):
        """Get the headers returned with the current page."""
        return self._assert_result_for('headers').headers

    @property
    def agent_string(self) -> Optional[str]:
        """Get the user agent string."""
        return self.headers.get('User-Agent')

    @agent_string.setter
    def agent_string(self, agent: str) -> None:
        """Set the user agent string to the given value."""
        self.headers['User-Agent'] = agent

    def show_forms(self) -> None:
        """Pretty-print all forms on the page.

        Include the global form (form elements outside <form> pairs)
        as forms[0] if present.
        """
        for n, form in enumerate(self.forms, 1):
            print_form(form, n)

    def show_links(self) -> None:
        """Pretty-print all links on the page."""
        info = log.info
        links = self.links
        if links:
            info('\nLinks (%d links total):\n', len(links))
            for n, link in enumerate(links, 1):
                info('\t%d. %s ==> %s', n, trunc(link.text, 40), link.url)
            info('')
        else:
            info('\n** no links **\n')

    def show_history(self) -> None:
        """Pretty-print the history of links visited."""
        info = log.info
        history = self._history
        if history:
            info('\nHistory (%d pages total):\n', len(history))
            for n, page in enumerate(history, 1):
                info('\t%d. %s', n, page.url)
            info('')
        else:
            info('\n** no history **\n')

    @property
    def links(self) -> List[Link]:
        """Return a list of all links on the page."""
        return self._assert_result_for('links').links

    @property
    def history(self) -> List[ResultWrapper]:
        """Return a list of all pages visited by the browser."""
        return self._history

    @property
    def forms(self) -> List[FormElement]:
        """Return a list of forms on the page.

        This includes the global form at index 0 if present.
        """
        return self._assert_result_for('forms').forms

    def form(self, name: Union[str, int] = 1) -> Optional[FormElement]:
        """Return the first form that matches the given form name."""
        return self._assert_result_for('form').form(name)

    def form_field(self, form: FormElement = None,
                   name_or_num: Union[str, int] = 1) -> FieldElement:
        """Return the control that matches the given field name.

        Must be a *unique* regex/exact string match, but the returned
        control can also be a CheckboxGroup or RadioGroup list.

        Raises a TwillException if no such field or multiple fields are found.
        """
        if form is None:
            form = self._form
            if form is None:
                raise TwillException("Must specify a form for the field")
        inputs = form.inputs
        found_multiple = False

        name = name_or_num if isinstance(name_or_num, str) else None

        if name:

            if name in form.fields:
                match_name = [c for c in inputs if c.name == name]
                if len(match_name) > 1:
                    if all(getattr(c, 'type', None) == 'checkbox'
                            for c in match_name):
                        return CheckboxGroup(
                            cast(List[InputElement], match_name))
                    if all(getattr(c, 'type', None) == 'radio'
                            for c in match_name):
                        return RadioGroup(cast(List[InputElement], match_name))
            else:
                match_name = None

            # test exact match to id
            match_id = [c for c in inputs if c.get('id') == name]
            if match_id:
                if unique_match(match_id):
                    return match_id[0]
                found_multiple = True

            # test exact match to name
            if match_name:
                if unique_match(match_name):
                    return match_name[0]
                found_multiple = True

        num = name_or_num if isinstance(name_or_num, int) else None
        if num is None and name and name.isdigit():
            try:
                num = int(name)
            except ValueError:
                pass

        # test field index
        if num is not None:
            try:
                return list(inputs)[num - 1]
            except IndexError:
                pass

        if name:

            # test regex match
            regex = re.compile(name)
            match_name = [c for c in inputs
                          if c.name and regex.search(c.name)]
            if match_name:
                if unique_match(match_name):
                    return match_name[0]
                found_multiple = True

            # test field values
            match_value = [c for c in inputs if c.value == name]
            if match_value:
                if len(match_value) == 1:
                    return match_value[0]
                found_multiple = True

        # error out
        if found_multiple:
            raise TwillException(f'multiple matches to "{name_or_num}"')
        raise TwillException(f'no field matches "{name_or_num}"')

    def add_form_file(self, field_name: str, fp: IO) -> None:
        self._form_files[field_name] = fp

    def clicked(self, form: FormElement, control: FieldElement) -> None:
        """Record a 'click' in a specific form."""
        if self._form != form:
            # construct a function to choose a particular form;
            # select_form can use this to pick out a precise form.
            self._form = form
            self.last_submit_button = None
        # record the last submit button clicked.
        if getattr(control, 'type', None) in ('submit', 'image'):
            self.last_submit_button = cast(InputElement, control)

    def submit(self, field_name: Optional[Union[str, int]] = None,
               form_name: Optional[Union[str, int]] = None) -> None:
        """Submit the last or specified form using the given field."""
        forms = self.forms
        if not forms:
            raise TwillException("There are no forms on this page.")

        ctl: Optional[InputElement] = None

        form = self._form if form_name is None else self.form(form_name)
        if form is None:
            if len(forms) > 1:
                raise TwillException(
                    "There is more than one form on this page;"
                    " therefore you must specify a form explicitly"
                    " or select one (use 'fv') before submitting.")
            form = forms[0]

        action = form.action or ''
        if '://' not in action:
            form.action = urljoin(self.url, action)

        # no field name?  see if we can use the last submit button clicked...
        if field_name is None:
            if form is not self._form or self.last_submit_button is None:
                # get first submit button in form.
                submits = [c for c in form.inputs
                           if getattr(c, 'type', None) in ('submit', 'image')]
                if submits:
                    ctl = cast(InputElement, submits[0])
            else:
                ctl = self.last_submit_button
        else:
            # field name given; find it
            ctl = cast(InputElement, self.form_field(form, field_name))

        # now set up the submission by building the request object that
        # will be sent in the form submission.
        if ctl is None:
            log.debug('Note: submit without using a submit button')
        else:
            log.info(
                "Note: submit is using submit button:"
                " name='%s', value='%s'", ctl.get('name'), ctl.value)

        # Add referer information.  This may require upgrading the
        # request object to have an 'add_unredirected_header' function.
        # @BRT: For now, the referrer is always the current page
        # @CTB: this seems like an issue for further work.
        # Note: We do not set Content-Type from form.attrib.get('enctype'),
        # since Requests does a much better job at setting the proper one.
        headers = {'Referer': self.url}

        payload = form.form_values()
        if ctl is not None:
            name = ctl.get('name')
            if name:
                payload.append((name, ctl.value or ''))
        encoded_payload = self._encode_payload(payload)

        # now actually GO
        if form.method == 'POST':
            if self._form_files:
                r = self._session.post(
                    form.action, data=encoded_payload, headers=headers,
                    files=self._form_files)
            else:
                r = self._session.post(
                    form.action, data=encoded_payload, headers=headers)
        else:
            r = self._session.get(
                form.action, params=encoded_payload, headers=headers)

        self._form = None
        self._form_files.clear()
        self.last_submit_button = None
        if self.result is not None:
            self._history.append(self.result)
        self.result = ResultWrapper(r)

    def cookies(self) -> RequestsCookieJar:
        """Get all cookies from the current session."""
        return self._session.cookies

    def save_cookies(self, filename: str) -> None:
        """Save cookies into the given file."""
        with open(filename, 'wb') as f:
            pickle.dump(self._session.cookies, f)

    def load_cookies(self, filename: str) -> None:
        """Load cookies from the given file."""
        with open(filename, 'rb') as f:
            self._session.cookies = pickle.load(f)

    def clear_cookies(self) -> None:
        """Delete all the cookies."""
        self._session.cookies.clear()

    def show_cookies(self) -> None:
        """Pretty-print all the cookies."""
        info = log.info
        cookies = self._session.cookies
        n = len(cookies)
        if n:
            log.info('\nThere are %d cookie(s) in the cookie jar.\n', n)
            for n, cookie in enumerate(cookies, 1):
                info('\t%d. %s', n, cookie)
            info('')
        else:
            log.info('\nThere are no cookies in the cookie jar.\n', n)

    def decode(self, value: Union[bytes, str]):
        """Decode a value using the current encoding."""
        if isinstance(value, bytes) and self.encoding:
            value = value.decode(self.encoding)
        return value

    def xpath(self, path: str) -> List[HtmlElement]:
        """Evaluate an xpath expression."""
        return self._assert_result_for('xpath').xpath(path)

    def _encode_payload(
            self, payload: Sequence[Tuple[str, Union[str, bytes]]]
            ) -> List[Tuple[str, Union[str, bytes]]]:
        """Encode a payload with the current encoding if not utf-8."""
        encoding = self.encoding
        if not encoding or encoding.lower() in ('utf8', 'utf-8'):
            return list(payload)
        return [(name, val if isinstance(val, bytes) else val.encode(encoding))
                for name, val in payload]

    @staticmethod
    def _get_meta_refresh(
            response: Response) -> Tuple[Optional[int], Optional[str]]:
        """Get meta refresh interval and url from a response."""
        try:
            tree = html_to_tree(response.text)
        except ValueError:
            # may happen when there is an XML encoding declaration
            tree = html_to_tree(response.content)
        try:
            content = tree.xpath(  # "refresh" is case insensitive
                "//meta[translate(@http-equiv,'REFSH','refsh')="
                "'refresh'][1]/@content")[0]
            interval, url = content.split(';', 1)
            interval = int(interval)
            if interval < 0:
                raise ValueError
            url = url.strip().strip('"').strip().strip("'").strip()
            url = url.split('=', 1)
            if url[0].strip().lower() != 'url':
                raise IndexError
            url = url[1].strip().strip('"').strip().strip("'").strip()
        except (IndexError, ValueError):
            interval = url = None
        else:
            if '://' not in url:  # relative URL, adapt
                url = urljoin(response.url, url)
        return interval, url

    _re_basic_auth = re.compile('Basic realm="(.*)"', re.I)

    def _journey(self, func_name, *args, **_kwargs):
        """Execute the function with the given name and arguments.

        The name should be one of 'open', 'reload', 'back', or 'follow_link'.
        This method then runs that function with the given arguments and turns
        the results into a nice friendly standard ResultWrapper object, which
        is stored as self.result.

        (Idea stolen from Python Browsing Probe (PBP).)
        """
        self._form = None
        self._form_files.clear()
        self.last_submit_button = None

        if func_name == 'open':
            url = args[0]

        elif func_name == 'follow_link':
            url = args[0]
            try:
                url = url.url
            except AttributeError:
                pass  # this is already a url
            if '://' not in url and self.url:
                url = urljoin(self.url, url)

        elif func_name == 'reload':
            url = self.url

        elif func_name == 'back':
            try:
                self.result = self._history.pop()
                return
            except IndexError:
                raise TwillException
        else:
            raise TwillException(f"Unknown function {func_name!r}")

        r = self._session.get(url, verify=self.verify)
        if r.status_code == 401:
            header = r.headers.get('WWW-Authenticate')
            realm = self._re_basic_auth.match(header)
            if realm:
                realm = realm.group(1)
                auth = self._auth.get((url, realm)) or self._auth.get(url)
                if auth:
                    r = self._session.get(url, auth=auth, verify=self.verify)

        # handle redirection via meta refresh (not handled in requests)
        refresh_interval = get_equiv_refresh_interval()
        if refresh_interval:
            visited = set()  # break circular refresh chains
            while True:
                interval, url = self._get_meta_refresh(r)
                if not url:
                    break
                if interval >= refresh_interval:
                    (log.info if self.show_refresh else log.debug)(
                        'Meta refresh interval too long: %d', interval)
                    break
                if url in visited:
                    log.warning('Circular meta refresh detected!')
                    break
                (log.info if self.show_refresh else log.debug)(
                    'Meta refresh to new URL: %s', url)
                r = self._session.get(url)
                visited.add(url)

        if func_name in ('follow_link', 'open'):
            # If we're really reloading and just didn't say so, don't store
            if self.result is not None and self.result.url != r.url:
                self._history.append(self.result)

        self.result = ResultWrapper(r)


browser = TwillBrowser()  # the global twill browser instance

_disable_insecure_request_warnings()  # should not warn for HTTP requests



"""Used in test_shell, to test default command execution & extensions."""

flag = False

__all__ = ['flag_true', 'assert_flag']


def flag_true():
    global flag
    flag = True


def assert_flag():
    global flag
    assert flag



# This file is part of the twill source distribution.
#
# twill is an extensible scriptlet language for testing Web apps,
# available at https://github.com/twill-tools/twill.
#
# Copyright (c) 2005-2022
# by C. Titus Brown, Ben R. Taylor, Christoph Zwerschke et al.
#
# This program and all associated source code files are released under the
# terms of the MIT license; please see the included LICENSE file for more
# information, or go to https://opensource.org/licenses/mit-license.php.

"""twill web browsing and testing language and associated utilities.

A scripting system for automating web browsing.  Useful for testing
web pages or grabbing data from password-protected sites automatically.
"""

import logging
import sys
import os.path

__version__ = '3.1'

__url__ = 'https://github.com/twill-tools/twill'
__download_url__ = 'https://pypi.org/project/twill/'

__all__ = [
    'browser', 'execute_file', 'execute_string',
    'log', 'set_log_level', 'set_output', 'set_err_out',
    'twill_ext', 'TwillCommandLoop']


this_dir = os.path.dirname(__file__)
# Add extensions directory at the *end* of sys.path.
# This means that user extensions will take priority over twill extensions.
extensions = os.path.join(this_dir, 'extensions')
sys.path.append(extensions)

twill_ext = '.twill'  # file extension for twill scripts


log_levels = dict(
    CRITICAL=logging.CRITICAL,
    ERROR=logging.ERROR,
    WARNING=logging.WARNING,
    INFO=logging.INFO,
    DEBUG=logging.DEBUG,
    NOTSET=logging.NOTSET)

log = logging.getLogger(__name__)
handler = None

stdout, stderr = sys.stdout, sys.stderr


def set_log_level(level=None):
    """Set the logging level.

    If no level is passed, use INFO as logging level.
    """
    if level is None:
        level = logging.INFO
    if isinstance(level, str):
        level = log_levels[level.upper()]
    log.setLevel(level)


def set_output(stream=None):
    """Set the standard output.

    If no stream is passed, use standard output.
    """
    global handler
    if stream is None:
        stream = stdout
    if handler:
        log.removeHandler(handler)
    handler = logging.StreamHandler(stream)
    log.addHandler(handler)
    sys.stdout = stream


def set_err_out(stream=None):
    """Set the error output.

    If no stream is passed, use standard error.
    """
    if stream is None:
        stream = stderr
    sys.stderr = stream


def shutdown():
    """Shut down and flush the logging system."""
    sys.stdout.flush()
    sys.stderr.flush()
    logging.shutdown()


set_log_level()
set_output()


# a convenience function:
from .browser import browser  # noqa: ignore=E402

# the two core components of twill:
from .parse import execute_file, execute_string  # noqa: ignore=E402
from .shell import TwillCommandLoop  # noqa: ignore=E402

# initialize global dict
from . import namespaces  # noqa: ignore=E402
namespaces.init_global_dict()



"""twill multiprocess execution system."""

import sys
import os
import time

from optparse import OptionParser

from . import execute_file, set_log_level

from pickle import load, dump

# make sure that the current working directory is in the path
if '' not in sys.path:
    sys.path.append('')


def main():

    try:
        fork = os.fork
    except AttributeError:
        sys.exit('Error: Must use Unix to be able to fork processes.')

    parser = OptionParser()
    add = parser.add_option
    add('-u', '--url', nargs=1, action="store", dest="url",
        help="start at the given URL before each script")
    add('-n', '--number', nargs=1, action="store", dest="number",
        default=1, type="int",
        help="number of times to run the given script(s)")
    add('-p', '--processes', nargs=1, action="store",
        dest="processes", default=1, type="int",
        help="number of processes to execute in parallel")

    options, args = parser.parse_args()

    if not args:
        sys.exit('Error: Must specify one or more scripts to execute.')

    average_number = options.number // options.processes
    last_number = average_number + options.number % options.processes
    child_pids = []
    is_parent = True
    repeat = 0

    # start a bunch of child processes and record their pids in the parent
    for i in range(options.processes):
        pid = fork()
        if pid:
            child_pids.append(pid)
        else:
            repeat = average_number if i else last_number
            is_parent = False
            break

    # set the children up to run and record their stats
    failed = False

    if is_parent:

        time.sleep(1)

        total_time = total_exec = 0

        # iterate over all the child pids, wait until they finish,
        # and then sum statistics
        for child_pid in child_pids[:]:
            child_pid, status = os.waitpid(child_pid, 0)
            if status:  # failure
                print(f'[twill-fork parent: process {child_pid} FAILED:'
                      f' exit status {status}]')
                print('[twill-fork parent:'
                      ' (not counting stats for this process)]')
                failed = True
            else:  # record statistics, otherwise
                filename = '.status.%d' % (child_pid,)
                with open(filename) as fp:
                    this_time, n_executed = load(fp)
                os.unlink(filename)
                total_time += this_time
                total_exec += n_executed

        # summarize
        print('\n----\n')
        print(f'number of processes: {options.processes}')
        print(f'total executed: {total_exec}')
        print(f'total time to execute: {total_time:.2f} s')
        if total_exec:
            avg_time = 1000 * total_time / total_exec
            print(f'average time: {avg_time:.2f} ms')
        else:
            print('(nothing completed, no average!)')
        print()

    else:

        pid = os.getpid()
        print(f'[twill-fork: pid {pid} : executing {repeat} times]')

        start_time = time.time()

        set_log_level('warning')
        for i in range(repeat):
            for filename in args:
                execute_file(filename, initial_url=options.url)

        end_time = time.time()
        this_time = end_time - start_time

        # write statistics
        filename = f'.status.{pid}'
        with open(filename, 'w') as fp:
            info = (this_time, repeat)
            dump(info, fp)

    sys.exit(-1 if failed else 0)


if __name__ == '__main__':
    main()



"""
A simple set of extensions to manage post-load requirements for pages.

Commands:

   require       -- turn on post-load requirements; either 'success' or
                    'links_ok'.

   no_require    -- turn off requirements.

   skip_require  -- for the next page visit, skip requirements processing.

   flush_visited -- flush the list of already visited pages
                    (for links checking)
"""

from twill import browser, commands, log

__all__ = ['require', 'skip_require', 'flush_visited', 'no_require']

_requirements = []      # what requirements to satisfy

ignore_once = False     # reset after each hook call
ignore_always = False   # never reset


def skip_require():
    """
    >> skip_require

    Skip the post-page-load requirements.
    """
    global ignore_once
    ignore_once = True


def require(what):
    """>> require <what>

    After each page is loaded, require that 'what' be satisfied.  'what'
    can be:
      * 'success' -- HTTP return code is 200
      * 'links_ok' -- all of the links on the page load OK (see 'check_links'
                      extension module)
    """
    global _requirements

    # install the post-load hook function.
    # noinspection PyProtectedMember
    hooks = browser._post_load_hooks
    if _require_post_load_hook not in hooks:
        log.debug('INSTALLING POST-LOAD HOOK')
        hooks.append(_require_post_load_hook)

    # add the requirement.
    if what not in _requirements:
        log.debug('Adding requirement: %s', what)
        _requirements.append(what)


def no_require():
    """>> no_require

    Remove all post-load requirements.
    """
    # noinspection PyProtectedMember
    hooks = browser._post_load_hooks
    hooks = [fn for fn in hooks if fn != _require_post_load_hook]
    browser._post_load_hooks = hooks

    global _requirements
    _requirements = []


def flush_visited():
    """>> flush_visited

    Flush the list of pages successfully visited already.
    """
    from .check_links import good_urls  # type: ignore
    good_urls.clear()


def _require_post_load_hook(action, *_args, **_kwargs):
    """Post load hook function to be called after each page is loaded.

    See TwillBrowser._journey() for more information.
    """
    if action == 'back':  # do nothing on a 'back'
        return

    global ignore_once
    global ignore_always

    if ignore_once or ignore_always:
        ignore_once = False
        return

    for what in _requirements:

        if what == 'success':
            log.debug('REQUIRING success')
            commands.code("200")

        elif what == 'links_ok':
            from check_links import check_links, good_urls  # type: ignore

            ignore_always = True
            log.debug('REQUIRING functioning links')
            log.debug('(already visited:)')
            log.debug("\n\t".join(sorted(good_urls)))

            try:
                check_links()
            finally:
                ignore_always = False



"""Code parsing and evaluation for the twill mini-language."""

import re
import sys

from io import StringIO
from typing import List

from pyparsing import (
    CharsNotIn, Combine, Group, Literal, Optional, ParseException,
    pyparsing_unicode, removeQuotes, restOfLine, Word, ZeroOrMore)

# noinspection PyCompatibility
from . import commands, log, namespaces
from .browser import browser
from .errors import TwillNameError

# pyparsing stuff

# allow characters in full 8bit range
char_range = pyparsing_unicode.Latin1
alphas, alphanums = char_range.alphas, char_range.alphanums
printables = char_range.printables

# basically, a valid Python identifier:
command_word = Word(alphas + '_', alphanums + '_')
command = command_word.setResultsName('command')
command.setName('command')

# arguments to it.

# we need to reimplement all this junk from pyparsing because pcre's
# idea of escapable characters contains a lot more than the C-like
# thing pyparsing implements
_bslash = '\\'
_sglQuote = Literal("'")
_dblQuote = Literal('"')
_escapables = printables
_escapedChar = Word(_bslash, _escapables, exact=2)
dblQuotedString = Combine(
    _dblQuote + ZeroOrMore(CharsNotIn('\\"\n\r') | _escapedChar | '""') +
    _dblQuote).streamline().setName("string enclosed in double quotes")
sglQuotedString = Combine(
    _sglQuote + ZeroOrMore(CharsNotIn("\\'\n\r") | _escapedChar | "''") +
    _sglQuote).streamline().setName('string enclosed in single quotes')
quotedArg = (dblQuotedString | sglQuotedString)
quotedArg.setParseAction(removeQuotes)
quotedArg.setName('quotedArg')

plainArgChars = printables.replace('#', '').replace('"', '').replace("'", "")
plainArg = Word(plainArgChars)
plainArg.setName('plainArg')

arguments_group = Group(ZeroOrMore(quotedArg | plainArg))
arguments = arguments_group.setResultsName('arguments')
arguments.setName('arguments')

# comment line.
comment = Literal('#') + restOfLine
comment = comment.suppress()
comment.setName('comment')

full_command = comment | (command + arguments + Optional(comment))
full_command.setName('full_command')


command_list: List[str] = []  # filled in by namespaces.init_global_dict().


def process_args(args, globals_dict, locals_dict):
    """Process string arguments.

    Take a list of string arguments parsed via pyparsing and evaluate
    the special variables ('__*').

    Return a new list.
    """
    new_args: List[str] = []
    for arg in args:
        # __variable substitution
        if arg.startswith('__'):
            try:
                val = eval(arg, globals_dict, locals_dict)
            except NameError:  # not in dictionary; don't interpret
                val = arg

            log.info('VAL IS %s FOR %s', val, arg)

            if isinstance(val, str):
                new_args.append(val)
            else:
                new_args.extend(val)

        # $variable substitution
        elif arg.startswith('$') and not arg.startswith('${'):
            try:
                val = str(eval(arg[1:], globals_dict, locals_dict))
            except NameError:  # not in dictionary; don't interpret
                val = arg
            new_args.append(val)
        else:
            new_args.append(variable_substitution(
                arg, globals_dict, locals_dict))

    new_args = [arg.replace('\\n', '\n') for arg in new_args]
    return new_args


def execute_command(cmd, args, globals_dict, locals_dict, cmdinfo):
    """Actually execute the command.

    Side effects: __args__ is set to the argument tuple, __cmd__ is set to
    the command.
    """
    global command_list  # all supported commands
    # execute command
    locals_dict['__cmd__'] = cmd
    locals_dict['__args__'] = args
    if cmd not in command_list:
        raise TwillNameError(f"unknown twill command: '{cmd}'")

    eval_str = f"{cmd}(*__args__)"

    # compile the code object so that we can get 'cmdinfo' into the
    # error tracebacks
    codeobj = compile(eval_str, cmdinfo, 'eval')

    # eval the codeobj in the appropriate dictionary
    result = eval(codeobj, globals_dict, locals_dict)

    # set __url__
    locals_dict['__url__'] = browser.url

    return result


_log_commands = log.debug  # type: ignore


def parse_command(line, globals_dict, locals_dict):
    """Parse command."""
    try:
        res = full_command.parseString(line)
    except ParseException as e:
        log.error('PARSE ERROR: %s', e)
        res = None
    if res:
        _log_commands("twill: executing cmd '%s'", line.strip())
        args = process_args(res.arguments.asList(), globals_dict, locals_dict)
        return res.command, args
    return None, None  # e.g. a comment


def execute_string(buf, **kw):
    """Execute commands from a string buffer."""
    fp = StringIO(buf)

    kw['source'] = ['<string buffer>']
    if 'no_reset' not in kw:
        kw['no_reset'] = True

    _execute_script(fp, **kw)


def execute_file(filename, **kw):
    """Execute commands from a file."""
    inp = sys.stdin if filename == '-' else open(filename, encoding='utf-8')

    log.info('\n>> Running twill file %s', filename)

    kw['source'] = filename
    _execute_script(inp, **kw)


def _execute_script(inp, **kw):
    """Execute lines taken from a file-like iterator."""
    # initialize new local dictionary and get global and current local
    namespaces.new_local_dict()
    globals_dict, locals_dict = namespaces.get_twill_glocals()

    locals_dict['__url__'] = browser.url

    # reset browser
    if not kw.get('no_reset'):
        commands.reset_browser()

    # go to a specific URL?
    init_url = kw.get('initial_url')
    if init_url:
        commands.go(init_url)
        locals_dict['__url__'] = browser.url

    # should we catch exceptions on failure?
    catch_errors = kw.get('never_fail')

    # source_info stuff
    source_info = kw.get('source', "<input>")

    try:

        for n, line in enumerate(inp, 1):
            line = line.strip()
            if not line:  # skip empty lines
                continue

            cmd_info = f'{source_info}:{n}'
            log.info('AT LINE: %s', cmd_info)

            cmd, args = parse_command(line, globals_dict, locals_dict)
            if cmd is None:
                continue

            try:
                execute_command(cmd, args, globals_dict, locals_dict, cmd_info)
            except SystemExit:
                # abort script execution if a SystemExit is raised
                return
            except Exception as e:
                error_type = e.__class__.__name__ or 'Error'
                error = f"{error_type} raised on line {n} of '{source_info}'"
                if line:
                    error += f" while executing\n>> {line}"
                log.error("\nOops! %s", error)
                if not browser.first_error:
                    browser.first_error = error
                log.error("\nError: %s", str(e).strip())
                if not catch_errors:
                    raise

    finally:
        cleanups = locals_dict.get('__cleanups__')
        if cleanups:
            error = browser.first_error
            result = browser.result
            for filename in reversed(cleanups):
                log.info('\n>> Running twill cleanup file %s', filename)
                try:
                    inp = open(filename, encoding='utf-8')
                    _execute_script(inp, source=filename, no_reset=True)
                except Exception as e:
                    log.error('>> Cannot run cleanup file %s: %s', filename, e)
            browser.reset()
            browser.first_error = error
            browser.result = result
        namespaces.pop_local_dict()


def log_commands(flag):
    """Turn printing of commands as they are executed on or off."""
    global _log_commands
    old_flag = _log_commands is log.info
    _log_commands = log.info if flag else log.debug
    return old_flag


_re_variable = re.compile(r"\${(.*?)}")


def variable_substitution(raw_str, globals_dict, locals_dict):
    s = []
    pos = 0
    for m in _re_variable.finditer(raw_str):
        s.append(raw_str[pos:m.start()])
        try:
            s.append(str(eval(m.group(1), globals_dict, locals_dict)))
        except NameError:
            s.append(m.group())
        pos = m.end()
    s.append(raw_str[pos:])
    return ''.join(s)



"""twill exceptions"""


class TwillException(Exception):
    """General twill exception."""
    pass


class TwillAssertionError(TwillException):
    """AssertionError to raise upon failure of some twill command."""
    pass


class TwillNameError(TwillException):
    """Error to raise when an unknown command is called."""
    pass



"""
Extension functions for manipulating the current working directory (cwd).

Commands:

   chdir -- push the cwd onto the directory stack & change to the new location.
   popd  -- change to the last directory on the directory stack.
"""

import os

from twill import commands, log

__all__ = ['chdir', 'popd']

_dir_stack = []


def chdir(where):
    """>> chdir <where>

    Change to the new location, after saving the current directory onto
    the directory stack.  The global variable __dir__ is set to the cwd.
    """
    cwd = os.getcwd()
    _dir_stack.append(cwd)
    log.debug('current directory: "%s"', cwd)

    os.chdir(where)
    log.info('changed directory to "%s"', where)

    commands.setglobal('__dir__', where)


def popd():
    """>> popd

    Change back to the last directory on the directory stack.  The global
    variable __dir__ is set to the cwd.
    """
    where = _dir_stack.pop()
    os.chdir(where)
    log.info('popped back to directory "%s"', where)

    commands.setglobal('__dir__', where)



"""
A command-line interpreter for twill.

This is an implementation of a command-line interpreter based on the
'Cmd' class in the 'cmd' package of the default Python distribution.
"""

import os
import sys
import traceback

from cmd import Cmd
from io import TextIOWrapper
from optparse import OptionParser
from typing import Any, Callable, List, Optional

try:
    from readline import read_history_file, write_history_file  # type: ignore
except ImportError:
    read_history_file = write_history_file = None  # type: ignore
# noinspection PyCompatibility
from . import (
    commands, execute_file,
    log, log_levels, set_log_level, set_output,
    namespaces, parse, shutdown, __url__, __version__)
from .browser import browser
from .utils import gather_filenames, Singleton

__all__ = ['main']

python_version = sys.version.split(None, 1)[0]

version_info = f"""
twill version: {__version__}
Python Version: {python_version}

See {__url__} for more info.
"""


def make_cmd_fn(cmd: str) -> Callable[[str], None]:
    """Make a command function.

    Dynamically define a twill shell command function based on an imported
    function name.  (This is where the twill.commands functions actually
    get executed.)
    """

    def do_cmd(rest_of_line: str, cmd: str = cmd) -> None:
        global_dict, local_dict = namespaces.get_twill_glocals()

        args = []
        if rest_of_line.strip():
            try:
                args = parse.arguments.parseString(rest_of_line)[0]
                args = parse.process_args(args, global_dict, local_dict)
            except Exception as e:
                log.error('\nINPUT ERROR: %s\n', e)
                return

        try:
            parse.execute_command(
                cmd, args, global_dict, local_dict, '<shell>')
        except SystemExit:
            raise
        except Exception as e:
            log.error('\nERROR: %s\n', e)
            log.debug(traceback.format_exc())

    return do_cmd


def make_help_cmd(cmd: str, docstring: str) -> Callable[[str], None]:
    """Make a help command function.

    Dynamically define a twill shell help function for the given
    command/docstring.
    """
    def help_cmd(message: str = docstring, cmd: str = cmd) -> None:
        message = message.strip()
        width = 7 + len(cmd)
        for line in message.splitlines():
            w = len(line.rstrip())
            if w > width:
                width = w
        info = log.info
        info('\n%s' % ('=' * width,))
        info('\nHelp for command %s:\n', cmd)
        info(message)
        info('\n%s\n' % ('=' * width,))

    return help_cmd


def add_command(cmd: str, docstring: str) -> None:
    """Add a command with given docstring to the shell."""
    shell = get_command_shell()
    if shell:
        shell.add_command(cmd, docstring)


class TwillCommandLoop(Singleton, Cmd):
    """The command-line interpreter for twill commands.

    This is a Singleton object: you can't create more than one
    of shell at a time.

    Note: most of the do_ and help_ functions are dynamically created
    by the metaclass.
    """

    def __init__(
            self,  stdin: Optional[TextIOWrapper] = None,
            initial_url: Optional[str] = None,
            fail_on_unknown: bool = False) -> None:
        Cmd.__init__(self, stdin=stdin)

        self.use_rawinput = stdin is None

        # initialize a new local namespace.
        namespaces.new_local_dict()

        # import readline history, if available/possible.
        if read_history_file:
            try:
                read_history_file('.twill-history')
            except IOError:
                pass

        # fail on unknown commands? for test-shell, primarily.
        self.fail_on_unknown = fail_on_unknown

        # handle initial URL argument
        if initial_url:
            commands.go(initial_url)

        self._set_prompt()

        self.names: List[str] = []

        global_dict, local_dict = namespaces.get_twill_glocals()

        # add all of the commands from twill
        for command in parse.command_list:
            fn = global_dict.get(command)
            self.add_command(command, fn.__doc__)

    def add_command(self, command: str, docstring: str) -> None:
        """Add the given command into the lexicon of all commands."""
        do_name = f'do_{command}'
        do_cmd = make_cmd_fn(command)
        setattr(self, do_name, do_cmd)

        if docstring:
            help_cmd = make_help_cmd(command, docstring)
            help_name = f'help_{command}'
            setattr(self, help_name, help_cmd)

        self.names.append(do_name)

    def get_names(self) -> List[str]:
        """Return the list of commands."""
        return self.names

    def complete_form_value(
            self, text: str, line: str, _begin: int, _end: int) -> List[str]:
        """Command arg completion for the form_value command.

        The twill command has the following syntax:

        form_value <form_name> <field_name> <value>
        """
        cmd, args = parse.parse_command(line + '.', {}, {})
        place = len(args)
        if place == 1:
            return self.provide_form_name(text)
        if place == 2:
            form_name = args[0]
            return self.provide_field_name(form_name, text)
        return []

    complete_fv = complete_form_value  # alias

    @staticmethod
    def provide_form_name(prefix: str) -> List[str]:
        """Provide the list of form names on the given page."""
        names = []
        forms = browser.forms
        for form in forms:
            form_id = form.attrib.get('id')
            if form_id and form_id.startswith(prefix):
                names.append(form_id)
                continue
            name = form.attrib.get('name')
            if name and name.startswith(prefix):
                names.append(name)
        return names

    @staticmethod
    def provide_field_name(form_name: str, prefix: str) -> List[str]:
        """Provide the list of fields for the given form_name or number."""
        names = []
        form = browser.form(form_name)
        if form is not None:
            for field in form.inputs:
                field_id = field.attrib.get('id')
                if field_id and field_id.startswith(prefix):
                    names.append(field_id)
                    continue
                name = field.name
                if name and name.startswith(prefix):
                    names.append(name)
        return names

    def _set_prompt(self) -> None:
        """"Set the prompt to the current page."""
        url = browser.url
        if url is None:
            url = " *empty page* "
        self.prompt = f"current page: {url}\n>> "

    def precmd(self, line: str) -> str:
        """Run before each command; save."""
        return line

    def postcmd(self, stop: bool, line: str) -> bool:
        """"Run after each command; set prompt."""
        self._set_prompt()
        return stop

    def default(self, line: str) -> None:
        """"Called when an unknown command is executed."""

        # empty lines ==> emptyline(); here we just want to remove
        # leading whitespace.
        line = line.strip()

        # look for command
        global_dict, local_dict = namespaces.get_twill_glocals()
        cmd, args = parse.parse_command(line, global_dict, local_dict)

        # ignore comments & empty stuff
        if cmd is None:
            return

        try:
            parse.execute_command(
                cmd, args, global_dict, local_dict, '<shell>')
        except SystemExit:
            raise
        except Exception as e:
            log.error('\nERROR: %s\n', e)
            if self.fail_on_unknown:
                raise

    def emptyline(self) -> Any:
        """Handle empty lines (by ignoring them)."""
        pass

    @staticmethod
    def do_EOF(*_args: str) -> None:
        """Exit on CTRL-D"""
        if write_history_file:
            write_history_file('.twill-history')
        raise SystemExit()

    @staticmethod
    def help_help() -> None:
        """Show help for the help command."""
        log.info("\nWhat do YOU think the command 'help' does?!?\n")

    @staticmethod
    def do_version(*_args: str) -> None:
        """Show the version number of twill."""
        log.info(version_info)

    @staticmethod
    def help_version() -> None:
        """Show help for the version command."""
        log.info("\nPrint version information.\n")

    def do_exit(self, *_args: str) -> None:
        """Exit the twill shell."""
        raise SystemExit()

    @staticmethod
    def help_exit() -> None:
        """Show help for the exit command."""
        log.info("\nExit twill.\n")

    do_quit = do_exit
    help_quit = help_exit


def get_command_shell() -> Optional[TwillCommandLoop]:
    """Get the command shell."""
    return getattr(TwillCommandLoop, '__it__', None)


twill_args: List[str] = []  # contains sys.argv *after* last '--'
interactive = False  # 'True' if interacting with user


def main():
    global twill_args, interactive

    # show the shorthand name for usage
    if sys.argv[0].endswith('-script.py'):
        sys.argv[0] = sys.argv[0].rsplit('-', 1)[0]

    # make sure that the current working directory is in the path
    if '' not in sys.path:
        sys.path.append('')

    parser = OptionParser()
    add = parser.add_option

    add('-d', '--dump-html', action='store', dest='dumpfile',
        help="dump HTML to this file on error")
    add('-f', '--fail', action='store_true', dest='fail',
        help='fail exit on first file to fail')
    add('-i', '--interactive', action='store_true', dest='interactive',
        help='drop into an interactive shell (after running files)')
    add('-l', '--loglevel', nargs=1, action='store', dest='loglevel',
        help='set the logging level')
    add('-n', '--never-fail', action='store_true', dest='never_fail',
        help='continue executing scripts past errors')
    add('-o', '--output', nargs=1, action='store', dest='outfile',
        help="print log to output file")
    add('-q', '--quiet', action='store_true', dest='quiet',
        help='do not show normal output')
    add('-u', '--url', nargs=1, action='store', dest='url',
        help='start at the given URL before each script')
    add('-v', '--version', action='store_true', dest='show_version',
        help='show version information and exit')
    add('-w', '--show-error-in-browser', action='store_true',
        dest='show_browser', help="show dumped HTML in a web browser ")

    # parse arguments
    sys_args = sys.argv[1:]
    if '--' in sys_args:
        for last in range(len(sys_args) - 1, -1, -1):
            if sys_args[last] == '--':
                twill_args = sys_args[last + 1:]
                sys_args = sys_args[:last]
                break

    options, args = parser.parse_args(sys_args)

    if options.show_version:
        log.info(version_info)
        sys.exit(0)

    quiet = options.quiet
    show_browser = options.show_browser
    dump_file = options.dumpfile
    out_file = options.outfile
    log_level = options.loglevel
    interactive = options.interactive or not args

    if out_file:
        out_file = out_file.lstrip('=').lstrip() or None
        if out_file == '-':
            out_file = None

    if interactive and (quiet or out_file or dump_file or show_browser):
        sys.exit("Interactive mode is incompatible with -q, -o, -d and -w")

    if options.show_browser and (not dump_file or dump_file == '-'):
        sys.exit("Please also specify a dump file with -d")

    if out_file:
        try:
            out_file = open(out_file, 'w')
        except IOError as e:
            sys.exit(f"Invalid output file '{out_file}': {e}")

    if log_level:
        log_level = log_level.lstrip('=').lstrip() or None
        if log_level.upper() not in log_levels:
            log_level_names = ', '.join(sorted(log_levels))
            sys.exit(f"Valid log levels are: {log_level_names}")
        set_log_level(log_level)

    if quiet:
        out_file = open(os.devnull, 'w')

    set_output(out_file)

    # first find and run any scripts put on the command line

    failed = False
    if args:
        success = []
        failure = []

        filenames = gather_filenames(args)
        dump = None

        for filename in filenames:
            try:
                interactive = False
                execute_file(filename, initial_url=options.url,
                             never_fail=options.never_fail)
                success.append(filename)
            except Exception as e:
                if dump_file:
                    dump = browser.dump
                if options.fail:
                    raise
                else:
                    if browser.first_error:
                        log.error('\nFirst error: %s', browser.first_error)
                    log.error('\n*** ERROR: %s', e)
                    log.debug(traceback.format_exc())
                    failure.append(filename)

        log.info('--')
        if dump:
            if dump_file == '-':
                log.info('HTML when error was encountered:\n\n%s\n--',
                         dump.strip())
            else:
                try:
                    with open(dump_file, 'wb') as f:
                        f.write(dump)
                except IOError as e:
                    log.error('Could not dump to %s: %s\n', dump_file, e)
                else:
                    log.info('HTML has been dumped to %s\n', dump_file)

        log.info('%d of %d files SUCCEEDED.',
                 len(success), len(success) + len(failure))
        if len(failure):
            log.error('Failed:\n\t%s', '\n\t'.join(failure))
            failed = True

        if dump and show_browser:
            import webbrowser

            url = 'file:///' + os.path.abspath(dump_file).replace(os.sep, '/')
            log.debug('Running web browser on %s', url)
            webbrowser.open(url)

    # if no scripts to run or -i is set, drop into an interactive shell

    if interactive:
        welcome_msg = "" if args else "\n -= Welcome to twill =-\n"

        shell = TwillCommandLoop(initial_url=options.url)

        while True:
            try:
                shell.cmdloop(welcome_msg)
            except KeyboardInterrupt:
                print()
                break
            except SystemExit:
                raise

            welcome_msg = ""

    shutdown()

    if failed:
        sys.exit(1)
    sys.exit(0)


if __name__ == '__main__':
    main()



"""Support functionality for using twill in unit tests."""

import sys
import time

from io import StringIO
from multiprocessing import Process

from .parse import execute_file

HOST = '127.0.0.1'  # interface to run the server on
PORT = 8080  # default port to run the server on
SLEEP = 0  # time to wait for the server to start


class TestInfo:
    """Test info container.

    Object containing info for a test: script to run, server function to
    run, and port to run it on.  Note that information about server port
    *must* be decided by the end of the __init__ function.

    The optional sleep argument specifies how many seconds to wait for the
    server to set itself up.  Default is 0.
    """

    def __init__(self, script, server_fn, port=PORT, sleep=SLEEP):
        self.script = script
        self.server_fn = server_fn
        self.port = port
        self.stdout = None
        self.stderr = None
        self.sleep = sleep

    def start_server(self):
        # save old stdout/stderr
        stdout, stderr = sys.stdout, sys.stderr

        # create new stdout/stderr
        self.stdout = sys.stdout = StringIO()
        self.stderr = sys.stderr = StringIO()

        try:
            self.server_fn()
        finally:
            # restore stdout/stderr
            sys.stdout, sys.stderr = stdout, stderr

    def run_script(self):
        """Run the given twill script on the given server."""
        time.sleep(self.sleep)
        url = self.url
        execute_file(self.script, initial_url=url)

    @property
    def url(self):
        """"Get the test server URL."""
        # noinspection HttpUrlsUsage
        return f"http://{HOST}:{self.port}/"


def run_test(test_info):
    """Run test on a web site where the site is running in a sub process."""
    # run server
    server_process = Process(target=test_info.start_server)
    server_process.start()
    # wait for server process to spin up
    timeout = max(1, test_info.sleep)
    wait = min(0.125, 0.125 * timeout)
    waited = 0
    while not server_process.is_alive() and waited < timeout:
        time.sleep(wait)
        waited += wait
    # run twill test script
    try:
        test_info.run_script()
    finally:
        server_process.terminate()



"""Global and local dictionaries, and initialization/utility functions."""

global_dict = {}  # the global dictionary


def init_global_dict():
    """Initialize the global dictionary with twill commands.

    This must be done after all the other modules are loaded, so that all
    the commands are already defined.
    """
    # noinspection PyCompatibility
    from . import commands, parse

    cmd_list = commands.__all__
    global_dict.update((cmd, getattr(commands, cmd)) for cmd in cmd_list)
    parse.command_list.extend(cmd_list)


_local_dict_stack = []  # local dictionaries


def new_local_dict():
    """Initialize a new local dictionary & push it onto the stack."""
    d = {}
    _local_dict_stack.append(d)
    return d


def pop_local_dict():
    """Get rid of the current local dictionary."""
    return _local_dict_stack.pop()


def get_twill_glocals():
    """Return both global and current local dictionary."""
    global global_dict, _local_dict_stack
    assert global_dict is not None, "must initialize global namespace first!"
    if not _local_dict_stack:
        new_local_dict()
    return global_dict, _local_dict_stack[-1]
